/*
 * This file is part of the Prescribe to Postscript converter by
 * Frank Pilhofer <fp@informatik.uni-frankfurt.de>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * For non-commercial use ONLY.
 */

/*
 * font and raster image handling
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "p2p.h"

char * p2p_font_c = "$Id: fonts.c,v 1.5 1999/08/05 11:15:42 fp Exp $";


ldfont *
havefont (int thefont)
{
  ldfont * iter = ldfonts;

  while (iter) {
    if (iter->fontnum == thefont)
      return iter;
    iter = iter->next;
  }
  return NULL;
}

void
kyoldfc (ldfc * tc)
{
  ldfont * tf;
  ldfc * iter, * last;

  if (tc == NULL || tc->data == NULL)
    return;

  if ((tf = havefont (tc->fontnum)) == NULL) {
    if ((tf = (ldfont *) malloc (sizeof (ldfont))) == NULL) {
      fprintf (stderr, "oops: cannot allocate memory for font\n");
      return;
    }
    tf->type           = LDFC_FONT;
    tf->page           = -1;
    tf->fontnum        = tc->fontnum;
    tf->ldfc.rotation  = tc->rotation;
    tf->ldfc.chars     = tc;
    tf->sfa.sfa        = 0;
    tf->next           = ldfonts;
    ldfonts            = tf;
    tc->next = NULL;
  }
  else {
    iter = tf->ldfc.chars;
    last = NULL;

    while (iter) {
      if (iter->code == tc->code) {
	break;
      }
      last = iter;
      iter = iter->next;
    }

    if (iter) {
      if (last) {
	tc->next = iter->next;
	last->next = tc;
      }
      else {
	tc->next = iter->next;
	tf->ldfc.chars = tc;
      }
      free (iter->data);
      free (iter);
    }
    else {
      tc->next = tf->ldfc.chars;
      tf->ldfc.chars = tc;
    }
    tf->page = -1;
    if (tc->rotation != tf->ldfc.rotation) {
      tf->ldfc.rotation = 0;
    }
  }
}

void
kyodelf (int thefont)
{
  ldfont *fiter, *flast;
  ldfc *citer, *cnext;

  fiter = ldfonts;
  flast = NULL;

  while (fiter) {
    if (fiter->fontnum == thefont) {
      break;
    }
    flast = fiter;
    fiter = fiter->next;
  }

  if (!fiter) {
    return;
  }

  if (flast) {
    flast->next = fiter->next;
  }
  else {
    ldfonts = fiter->next;
  }

  citer = fiter->ldfc.chars;

  while (citer) {
    cnext = citer->next;
    free (citer->data);
    free (citer);
    citer = cnext;
  }

  free (fiter);
}

void
kyodaf (void)
{
  while (ldfonts) kyodelf (ldfonts->fontnum);
}

void
kyosfa (int thefont, double hmi, double vmi, int fp)
{
  ldfont *iter;

  if ((iter = havefont (thefont)) == NULL) {
    if ((iter = (ldfont *) malloc (sizeof (ldfont))) == NULL) {
      fprintf (stderr, "oops: cannot allocate font structure\n");
      return;
    }
    iter->page = -1;
    iter->fontnum = thefont;
    iter->ldfc.chars = NULL;
    iter->sfa.sfa = 1;
    iter->sfa.hmi = hmi;
    iter->sfa.vmi = vmi;
    iter->sfa.fp  = fp;

    iter->next = ldfonts;
    ldfonts = iter;
  }
  else {
    iter->sfa.sfa = 1;
    iter->sfa.hmi = hmi;
    iter->sfa.vmi = vmi;
    iter->sfa.fp  = fp;
  }
}

void
kyosfnt (char * name, double height, int fontno, int symbolset,
	 double compression, double shear)
{
  ldfont newfont, *iter, *next;
  int i;

  newfont.page = -1;
  newfont.type = SFNT_FONT;
  newfont.fontnum = fontno;
  newfont.sfa.sfa = 0;
  newfont.sfnt.height = height;
  newfont.sfnt.compression = compression;
  newfont.sfnt.shear = shear;

  for (i=1; i<MAXFONT; i++) {
    if (strcmp (name, fontmap[i].sfnt) == 0 && spo == fontmap[i].po) {
      newfont.sfnt.base = i;
      break;
    }
  }

  if (i == MAXFONT) {
    newfont.sfnt.base = 0;
    for (i=0; *scalable[i].font; i++) {
      if (strcmp (name, scalable[i].sfnt) == 0) {
	strcpy (newfont.sfnt.name, scalable[i].font);
	if (strstr (scalable[i].sfnt, "Condensed") != NULL) {
	  newfont.sfnt.compression *= 0.7;
	}
	break;
      }
    }

    if (!*scalable[i].font) {
      fprintf (stderr, "line %d: unknown font in sfnt %s\n", lineno, name);
      return;
    }

    if (height <= 0) {
      fprintf (stderr, "line %d: warning: sfnt default font size not supported\n",
	       lineno);
      height = 10;
    }
  }

  memcpy (&curfont, &newfont, sizeof (ldfont));

  if (fontno > 0) {
    if (fontno < 1000) {
      fprintf (stderr, "line %d: illegal font number %d in sfnt\n",
	       lineno, fontno);
    }
    else {
      if ((iter = havefont (fontno)) == NULL) {
	if ((iter = (ldfont *) malloc (sizeof (ldfont))) == NULL) {
	  fprintf (stderr, "oops: cannot allocate font structure\n");
	  return;
	}
	curfont.next = ldfonts;
	memcpy (iter, &curfont, sizeof (ldfont));
	ldfonts = iter;
      }
      else {
	curfont.next = iter->next;
	memcpy (iter, &curfont, sizeof (ldfont));
      }
    }
  }

  font = 0;
  kyofont ();
}

/*
 * Load Kyocera font as Postscript Type 3 font
 */

void
loadfont (int thefont)
{
  int chars, ix, iy, count, c;
  unsigned char *ptr;
  ldfont *tf;
  ldfc *iter;

  if (thefont == 0) {
    tf = &curfont;
  }
  else if ((tf = havefont (thefont)) == NULL) {
    return;
  }

  /*
   * Already defined on current page?
   */

  if (tf->page == page && thefont != 0)
    return;

  /*
   * Handle loaded font
   */

  if (tf->type == LDFC_FONT) {

    /*
     * On first use, determine bounding box
     */

    if (tf->page == -1) {
      tf->ldfc.minx =  10000.0;
      tf->ldfc.miny =  10000.0;
      tf->ldfc.maxx = -10000.0;
      tf->ldfc.maxy = -10000.0;
      
      for (iter=tf->ldfc.chars, chars=0; iter; iter=iter->next) {
	if (iter->data == NULL) {
	  continue;
	}
	
	chars++;
	
	if (iter->xoffset/32.0 < tf->ldfc.minx)
	  tf->ldfc.minx = iter->xoffset / 32.0 - iter->cellctr / 32.0;
	if (iter->yoffset/32.0 - iter->height < tf->ldfc.miny)
	  tf->ldfc.miny = iter->yoffset / 32.0 - iter->height;
	if (iter->xoffset/32.0 + iter->width > tf->ldfc.maxx)
	  tf->ldfc.maxx = iter->xoffset / 32.0 - iter->cellctr / 32.0 + iter->width;
	if (iter->yoffset/32.0 > tf->ldfc.maxy)
	  tf->ldfc.maxy = iter->yoffset / 32.0;
      }
    }
    
    tf->page = page;
    
    fprintf (outfile, "%% loading font %d\n", thefont);
    fprintf (outfile, "9 dict begin\n");
    fprintf (outfile, "/FontType 3 def\n");
    fprintf (outfile, "/FontMatrix [%lf 0 0 %lf 0 0] def\n",
	     scale * umumap[UNIT_PIXEL], -scale * umumap[UNIT_PIXEL]);
    fprintf (outfile, "/FontBBox [%lf %lf %lf %lf] def\n",
	     tf->ldfc.minx, tf->ldfc.miny, tf->ldfc.maxx, tf->ldfc.maxy);
    fprintf (outfile, "/Encoding 256 array def\n");
    fprintf (outfile, "0 1 255 {Encoding exch /.notdef put} for\n");
    fprintf (outfile, "/CharProcs %d dict def\n", chars+1);
    fprintf (outfile, "/CharMetrics %d dict def\n", chars+1);
    fprintf (outfile, "\
/BuildGlyph {\n\
  dup 2 index /CharMetrics get exch\n\
  2 copy known not {pop /.notdef} if\n\
  get exec setcachedevice\n\
  exch /CharProcs get exch\n\
  2 copy known not {pop /.notdef} if\n\
  get exec\n\
} bind def\n\
/BuildChar {\n\
  1 index /Encoding get exch get\n\
  1 index /BuildGlyph get exec\n\
} bind def\n\
");

    fprintf (outfile, "CharProcs /.notdef {} put\n");
    fprintf (outfile, "CharMetrics /.notdef {0 0 0 0 0 0} put\n");
    
    for (iter=tf->ldfc.chars; iter; iter=iter->next) {
      if (iter->data == NULL) {
	continue;
      }
      fprintf (outfile, "%% ldfc %d, %d\n", iter->fontnum, iter->code);
      fprintf (outfile, "Encoding %d /%d put\n", iter->code, iter->code);
      fprintf (outfile, "CharMetrics /%d {%lf %lf %lf %lf %lf %lf} put\n",
	       iter->code,
	       (double) iter->cellwidth / 32.0, 0.0,
	       (double) iter->xoffset / 32.0 - iter->cellctr / 32.0,
	       (double) iter->yoffset / 32.0 - iter->height,
	       (double) iter->xoffset / 32.0 - iter->cellctr / 32.0 + iter->width,
	       (double) iter->yoffset / 32.0);
      /*
       * produce appropriate character image.
       */
      fprintf (outfile, "CharProcs /%d {\n", iter->code);
      fprintf (outfile, "  %lf %lf moveto\n",
	       (double) iter->xoffset / 32.0 - iter->cellctr / 32.0,
	       (double) iter->yoffset / 32.0 - iter->height);
      fprintf (outfile, "  currentpoint translate %d %d scale\n",
	       iter->width, iter->height);
      fprintf (outfile, "  %d %d true [%d 0 0 %d 0 %d]\n",
	       iter->width, iter->height,
	       iter->width, -iter->height, iter->height);
      fprintf (outfile, "  <");
      
      ptr   = iter->data;
      ix    = iy = 0;
      count = 0;
      
      while (iy < iter->height) {
	c = (*ptr)>>4;   fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
	c = (*ptr)&0x0f; fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
	if ((++count % 30) == 0) {
	  fprintf (outfile, "\n   ");
	}
	
	ptr++;
	ix += 8;
	
	if (ix >= iter->width) {
	  if ((ix % 16) != 0) ptr++;
	  ix = 0;
	  iy++;
	}
      }
      fprintf (outfile, ">\n");
      fprintf (outfile, "  imagemask\n");
      fprintf (outfile, "} put\n");
    }

    fprintf (outfile, "currentdict end\n");
    fprintf (outfile, "/kyofont.%d exch definefont pop\n", thefont);
    fprintf (outfile, "%% end of font %d, %d chars loaded\n", thefont, chars);
    fprintf (outfile, "/kyofont.%d findfont 1 scalefont setfont\n",
	     thefont);
  }
  else if (tf->type == SFNT_FONT) {
    if (tf->sfnt.base != 0) {
      fprintf (outfile, "%lf %lf %lf %lf %d %lf /%s-ISOLatin1 /%s kyofont\n",
	       sls,
	       scs,
	       tf->sfnt.shear,
	       tf->sfnt.compression,
	       spo,
	       fontmap[tf->sfnt.base].size * scale,
	       fontmap[tf->sfnt.base].font,
	       fontmap[tf->sfnt.base].font);
    }
    else {
      fprintf (outfile, "%lf %lf %lf %lf %d %lf /%s-ISOLatin1 /%s kyofont\n",
	       sls,
	       scs,
	       tf->sfnt.shear,
	       tf->sfnt.compression,
	       spo,
	       tf->sfnt.height * scale,
	       tf->sfnt.name,
	       tf->sfnt.name);
    }
  }
}

/*
 * dump the loaded fonts as a C file
 */

void
dumpfonts (FILE *out)
{
  ldfont *fiter=ldfonts;
  ldfc *citer;
  int count, xi, yi;
  unsigned char *ptr;

  fprintf (out, "#include <stdio.h>\n");
  fprintf (out, "#include <stdlib.h>\n");
  fprintf (out, "#include \"p2p.h\"\n");
  
  while (fiter) {
    for (citer=fiter->ldfc.chars; citer; citer=citer->next) {
      if (citer->data == NULL) {
	continue;
      }
      fprintf (out, "static ldfc kyofont_%d_%d = {\n",
	       citer->fontnum, citer->code);
      fprintf (out, "  %d, %d, %d,\n", -1, citer->fontnum, citer->code);
      fprintf (out, "  %d, %d, %d, %d,\n",
	       citer->height, citer->width, citer->yoffset, citer->xoffset);
      fprintf (out, "  %d, %d, %d,\n",
	       citer->cellwidth, citer->cellctr, citer->rotation);

      ptr = citer->data;
      count = 0;

      for (yi=0; yi<citer->height; yi++) {
	for (xi=0; xi<citer->width; xi+=16) {
	  if (count % 16 == 0) {
	    fprintf (out, "  \"");
	  }
	  fprintf (out, "\\%03o\\%03o", (int) *ptr, (int) *(ptr+1));
	  ptr += 2;
	  count += 2;
	  if (count % 16 == 0) {
	    fprintf (out, "\"\n");
	  }
	}
      }

      if (count % 16) {
	fprintf (out, "\"\n");
      }
      fprintf (out, "  ,\n");
      fprintf (out, "  0, 0, NULL\n");
      fprintf (out, "};\n\n");
    }
    fiter = fiter->next;
  }

  fiter = ldfonts;

  fprintf (out, "void loadfonts (void)\n");
  fprintf (out, "{\n");

  while (fiter) {
    for (citer=fiter->ldfc.chars; citer; citer=citer->next) {
      if (citer->data == NULL) {
	continue;
      }
      fprintf (out, "  kyoldfc (&kyofont_%d_%d);\n",
	       citer->fontnum, citer->code);
    }
    if (fiter->sfa.sfa) {
      fprintf (out, "  kyosfa  (%d, %lf, %lf, %d);\n",
	       fiter->fontnum, fiter->sfa.hmi,
	       fiter->sfa.vmi, fiter->sfa.fp);
    }
    fiter = fiter->next;
  }
  fprintf (out, "}\n");
}

void
kyorvrd (rvrd * data)
{
  int xmax, ymax, xi, c;
  unsigned char *ptr;
  rvrd *iter;

  for (ymax=0, iter=data, xmax=0; iter; iter=iter->next, ymax++) {
    if (iter->length > xmax) {
      xmax = iter->length;
    }
  }

  fprintf (outfile, "%% loading raster image size %dx%d\n", xmax*8, ymax);
  fprintf (outfile, "matrix currentmatrix\n");
  fprintf (outfile, "1 spo eq 1 sro ne and {90 rotate} if\n");
  fprintf (outfile, "currentpoint translate\n");
  fprintf (outfile, "%lf %lf mul %lf %lf mul scale\n",
	   72.0 / (double) str, scale*xmax*8.0,
	   72.0 / (double) str, scale*ymax);
  fprintf (outfile, "%d %d true [%d 0 0 %d 0 %d]\n",
	   xmax*8, ymax, xmax*8, -ymax, ymax);
  fprintf (outfile, "<");

  for (iter=data; iter; iter=iter->next) {
    for (xi=0, ptr=iter->data; xi<xmax; xi++, ptr++) {
      if (xi < iter->length) {
	c = (*ptr)>>4;   fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
	c = (*ptr)&0x0f; fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
      }
      else {
	c = 0;           fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
	c = 0;           fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
      }
    }
    if (iter->next) fprintf (outfile, "\n ");
  }
  fprintf (outfile, ">\n");
  fprintf (outfile, "imagemask\n");
  fprintf (outfile, "setmatrix\n");
}

void
kyokillrd (rvrd * data)
{
  rvrd *next;

  while (data) {
    next = data->next;
    if (data->data) {
      free (data->data);
    }
    free (data);
    data = next;
  }
}

void
kyofpat (unsigned char *data)
{
  int xi, c;

  fprintf (outfile, "%% loading new fill pattern\n");
  fprintf (outfile, "/fpat <");

  for (xi=0; xi<32; xi++) {
    c = (*data)>>4;   fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
    c = (*data)&0x0f; fputc ((c>9)?(c-10+'a'):(c+'0'), outfile);
    data++;
  }

  fprintf (outfile, "> def\n");
}
