/*
 * This file is part of the Prescribe to Postscript converter by
 * Frank Pilhofer <fp@informatik.uni-frankfurt.de>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * For non-commercial use ONLY.
 */


#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>

#ifdef HAVE_PIPE
#include <unistd.h>
#include <sys/wait.h>
#endif

#include "p2p.h"

char * p2p_main = "$Id: main.c,v 1.17 2000/09/19 09:38:34 fp Exp $";

/*
 * Line counter for scanner
 */

int lineno;

/*
 * set 1 to merge kyo input with ps output
 */

int kyoout = 0;

/*
 * create document structuting comments
 */

int dsc = 1;

/*
 * from p2p.y
 */

extern int yydebug;

/*
 * input files
 */

char **inputfiles;
int inputfile;

/*
 * set up the postscript device
 */

void
ps_setup (FILE *thefile)
{
  time_t now;
  int iter;

  now = time (NULL);

  /*
   * Print Postscript Intro
   */

  if (dsc) {
    fprintf (thefile, "\
%%!PS-Adobe-3.0\n\
%%%%Creator:      Prescribe-to-Postscript converter by FP\n\
%%%%CreationDate: %s\
%%%%Title:       ",
	     ctime (&now));

    for (iter=0; inputfiles[iter]; iter++) {
      if (inputfiles[iter][0]=='-'&&!inputfiles[iter][1]) {
	fprintf (thefile, " (stdin)");
      }
      else {
	fprintf (thefile, " %s", inputfiles[iter]);
      }
    }

    fprintf (thefile, "\n\
%%%%EndComments\n\
%%%%BeginProlog\n\
");
  }
  else {
    fprintf (thefile, "%%!PS-Adobe\n");
  }

  /*
   * define some macros
   */

  fprintf (thefile, "/level2 true def\n");

  if (encoding == LASERJET) {
    fprintf (thefile, "\
/LaserJetEncoding [\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 00-07\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 08-0f\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 10-17\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 18-1f\n\
/space/exclam/quotedbl/numbersign/dollar/percent/ampersand/quoteright\n\
/parenleft/parenright/asterisk/plus/comma/minus/period/slash        %% 28-2f\n\
/zero/one/two/three/four/five/six/seven                             %% 30-37\n\
/eight/nine/colon/semicolon/less/equal/greater/question             %% 38-3f\n\
/at/A/B/C/D/E/F/G/H/I/J/K/L/M/N/O/P/Q/R/S/T/U/V/W                   %% 40-57\n\
/X/Y/Z/bracketleft/backslash/bracketright/asciicircum/underscore    %% 58-5f\n\
/quoteleft/a/b/c/d/e/f/g/h/i/j/k/l/m/n/o/p/q/r/s/t/u/v/w            %% 60-77\n\
/x/y/z/braceleft/bar/braceright/asciitilde/.notdef                  %% 78-7f\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 80-87\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 88-8f\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 90-97\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 98-9f\n\
/space/Agrave/Acircumflex/Egrave/Ecircumflex/Edieresis/Icircumflex/Idieresis\n\
/acute/grave/circumflex/dieresis/tilde/Ugrave/Ucircumflex/sterling  %% a8-af\n\
/macron/Yacute/yacute/degree/Ccedilla/ccedilla/Ntilde/ntilde        %% b0-b7\n\
/exclamdown/questiondown/currency/sterling/yen/section/florin/cent  %% b8-bf\n\
/acircumflex/ecircumflex/ocircumflex/ucircumflex/aacute/.notdef/oacute/uacute\n\
/agrave/egrave/ograve/ugrave/adieresis/edieresis/odieresis/udieresis%% c8-cf\n\
/Aring/icircumflex/Oslash/AE/aring/iacute/oslash/ae                 %% d0-d7\n\
/Adieresis/igrave/Odieresis/Udieresis                               %% d8-db\n\
/Eacute/idieresis/germandbls/Ocircumflex                            %% dc-df\n\
/Aacute/Atilde/atilde/Eth/eth/Iacute/Igrave/Oacute                  %% e0-e7\n\
/Ograve/Otilde/otilde/Scaron/scaron/Uacute/Ydieresis/ydieresis      %% e8-ef\n\
/Thorn/thorn/dotaccent/mu/paragraph/threequaters/emdash/onequarter  %% f0-f7\n\
/onehalf/ordfeminine/ordmasculine/guillemotleft                     %% f8-fb\n\
/.notdef/guillemotright/plusminus/.notdef                           %% fc-ff\n\
] def\n\
");
    fprintf (thefile, "/TheEncoding LaserJetEncoding def\n");
  }
  else if (encoding == IBMPC) {
    fprintf (thefile, "\
/IBMPCEncoding [\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 00-07\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 08-0f\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 10-17\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% 18-1f\n\
/space/exclam/quotedbl/numbersign/dollar/percent/ampersand/quoteright\n\
/parenleft/parenright/asterisk/plus/comma/minus/period/slash        %% 28-2f\n\
/zero/one/two/three/four/five/six/seven                             %% 30-37\n\
/eight/nine/colon/semicolon/less/equal/greater/question             %% 38-3f\n\
/at/A/B/C/D/E/F/G/H/I/J/K/L/M/N/O/P/Q/R/S/T/U/V/W                   %% 40-57\n\
/X/Y/Z/bracketleft/backslash/bracketright/asciicircum/underscore    %% 58-5f\n\
/quoteleft/a/b/c/d/e/f/g/h/i/j/k/l/m/n/o/p/q/r/s/t/u/v/w            %% 60-77\n\
/x/y/z/braceleft/bar/braceright/asciitilde/.notdef                  %% 78-7f\n\
/Ccedilla/udieresis/eacute/acircumflex/adieresis/agrave/aring/ccedilla\n\
/ecircumflex/edieresis/egrave/idieresis/icircumflex/igrave/Adieresis/Aring\n\
/Eacute/ae/AE/ocircumflex/odieresis/ograve/ucircumflex/ugrave       %% 90-97\n\
/ydieresis/Odieresis/Udieresis/cent/sterling/yen/paragraph/florin   %% 98-9f\n\
/aacute/iacute/oacute/uacute/ntilde/Ntilde/ordfeminine/ordmasculine %% a0-a7\n\
/questiondown/.notdef/.notdef/onehalf                               %% a8-ab\n\
/onequarter/exclamdown/guillemotleft/guillemotright                 %% ac-af\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% b0-b7\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% b8-bf\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% c0-c7\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% c8-cf\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% d0-d7\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% d8-df\n\
/.notdef/germandbls/.notdef/.notdef/.notdef/.notdef/mu/.notdef      %% e0-e7\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% e8-ef\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% f0-f7\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef    %% f8-ff\n\
] def\n\
");
    fprintf (thefile, "/TheEncoding IBMPCEncoding def\n");
  }
  else {
    fprintf (thefile, "/TheEncoding ISOLatin1Encoding def\n");
  }

  fprintf (thefile, "\
/setencoding {\n\
  level2\n\
  { findfont\n\
    dup length dict begin\n\
      {1 index /FID ne {def} {pop pop} ifelse} forall\n\
      /Encoding TheEncoding def\n\
      currentdict\n\
    end\n\
    definefont }\n\
  { findfont exch pop }\n\
  ifelse\n\
} def\n\
");

  /*
   * sls scs prop shear compression spo size font-ISOLatin1 font kyofont
   */

  fprintf (thefile, "\
/kyofont {\n\
  setencoding exch scalefont exch\n\
  dup ftmd 8 and 0 ne { dup /spo exch def } if pop\n\
  dup spo ne\n\
    { 1 eq { [ 0 1 -1 0 0 0 ] makefont } { [ 0 -1 1 0 0 0 ] makefont } ifelse }\n\
    { pop }\n\
    ifelse\n\
  [ 1 0 0 1 0 0 ] dup 0 5 -1 roll put makefont %% compression\n\
  [ 1 0 0 1 0 0 ] dup 2 5 -1 roll put makefont %% shear\n\
  [ 1 0 0 -1 0 0 ] makefont\n\
  setfont\n\
  ftmd 4 and 0 ne\n\
    { /prop exch def }\n\
    { pop }\n\
    ifelse\n\
  dup /rcs exch def ftmd 1 and 0 ne { dup /scs exch def } if pop\n\
  ftmd 2 and 0 ne { dup /sls exch def } if pop\n\
  prop 0 eq\n\
    { currentfont matrix dup 0 rcs scs 0\n\
      dup 0 ne { exch pop } { pop } ifelse\n\
      dup 0 eq { pop 1 } if\n\
      div put\n\
      makefont setfont }\n\
    if\n\
} def\n\
");

  fprintf (thefile, "\
/kyomove {\n\
  3 -1 roll\n\
  dup 0 eq\n\
  { pop pop pop }\n\
  { dup 1 eq\n\
    { pop pop 0 rmoveto }\n\
    { dup 2 eq\n\
      { pop 0 exch rmoveto pop }\n\
      { dup 3 eq\n\
        { pop rmoveto }\n\
        { dup 4 eq\n\
          { pop pop pop 0 sls rmoveto }\n\
          { pop pop pop currentpoint sls add 0 exch moveto pop }\n\
          ifelse }\n\
        ifelse }\n\
      ifelse }\n\
    ifelse }\n\
  ifelse\n\
} def\n\
");

  fprintf (thefile, "\
/underline {\n\
  currentpoint newpath moveto\n\
  currentpoint currentlinewidth 4 -1 roll\n\
  sulps setlinewidth\n\
  0 sulpd rmoveto\n\
  0 rlineto\n\
  stroke\n\
  setlinewidth\n\
  moveto\n\
} def\n\
");

  fprintf (thefile, "\
/box {\n\
  currentpoint newpath moveto\n\
  2 copy\n\
  dup 3 -1 roll dup 4 -1 roll\n\
  exch dup 0 rlineto\n\
  exch dup 0 exch rlineto\n\
  exch dup 0 exch sub 0 rlineto\n\
  pop 0 exch sub 0 exch rlineto\n\
  currentpoint closepath stroke moveto\n\
  pop pop kyomove\n\
} def\n\
");

  fprintf (thefile, "\
/block {\n\
  2 copy\n\
  matrix currentmatrix currentpoint 2 copy translate 5 -2 roll\n\
  2 copy scale\n\
  pixel div dup 0 lt { neg } if .5 add cvi exch\n\
  pixel div dup 0 lt { neg } if .5 add cvi\n\
  dup 1 string exch 0 0 6 -2 roll exch\n\
  2 copy matrix scale true exch\n\
  {\n\
    2 copy 16 mod 2 mul exch 16 mod 8 div cvi add fpat exch get\n\
    4 2 roll 8 add 2 copy le { pop 0 4 -1 roll 1 add } { 4 -1 roll } ifelse\n\
    4 -1 roll 4 index exch 0 exch put 3 index\n\
  }\n\
  imagemask\n\
  pop pop pop pop 3 -1 roll\n\
  setmatrix moveto kyomove\n\
} def\n\
");

  fprintf (thefile, "\
/cir {\n\
  currentpoint currentpoint 5 -1 roll\n\
  0 360 newpath arc stroke moveto\n\
} def\n\
");

  fprintf (thefile, "\
/kyoarc {\n\
  -90 add exch -90 add exch\n\
  currentpoint currentpoint currentpoint newpath\n\
  8 index 8 index 8 index arc\n\
  7 index 5 index 7 index arcn fill\n\
  moveto pop pop pop pop\n\
} def\n\
");

  fprintf (thefile, "\
/kyoshow {\n\
  prop 1 eq\n\
  { show }\n\
  {\n\
    currentpoint 3 -1 roll\n\
    { pop pop moveto scs 0 rmoveto currentpoint } exch kshow\n\
    pop pop\n\
  }\n\
  ifelse\n\
} def\n\
");

  fprintf (thefile, "\
/kyostringwidth {\n\
  prop 1 eq\n\
  { stringwidth }\n\
  { length scs mul 0 }\n\
  ifelse\n\
} def\n\
");

  fprintf (thefile, "\
/text {\n\
  currentpoint 3 -1 roll\n\
  dup kyoshow kyostringwidth\n\
  4 2 roll moveto\n\
  2 copy pop 4 -1 roll\n\
  0 ne { dup underline } if pop\n\
  kyomove\n\
} def\n\
");

  fprintf (thefile, "\
/ctxt {\n\
  currentpoint 3 -1 roll\n\
  dup kyostringwidth 2 copy\n\
  exch 0 exch sub 2 div exch rmoveto\n\
  currentpoint 5 -1 roll kyoshow moveto\n\
  2 copy pop 6 -1 roll\n\
  0 ne { dup underline } if pop\n\
  4 2 roll moveto\n\
  exch 2 div exch kyomove\n\
} def\n\
");

  fprintf (thefile, "\
/rtxt {\n\
  currentpoint 3 -1 roll\n\
  dup kyostringwidth 2 copy\n\
  exch 0 exch sub exch rmoveto\n\
  currentpoint 5 -1 roll kyoshow moveto\n\
  2 copy pop 6 -1 roll\n\
  0 ne { dup underline } if pop\n\
  4 2 roll moveto\n\
  exch 0 exch sub exch kyomove\n\
} def\n\
");

  fprintf (thefile, "\
/scp { \n\
   PosStackPos 8 lt { \n\
     xPosStack PosStackPos currentpoint pop put \n\
     yPosStack PosStackPos currentpoint exch pop put \n\
     /PosStackPos PosStackPos 1 add def \n\
   } if \n\
} def\n");
  fprintf (thefile, "\
/rpp { \n\
   PosStackPos 0 gt { \n\
     /PosStackPos PosStackPos 1 sub def \n\
     xPosStack PosStackPos get \n\
     yPosStack PosStackPos get \n\
     moveto \n\
   } if \n\
} def\n");

  if (dsc) fprintf (thefile, "%%%%EndProlog\n");

  /*
   * define positon stack
   */

  if (dsc) fprintf (thefile, "%%%%BeginSetup\n");

  fprintf (thefile, "/xPosStack 8 array def\n");
  fprintf (thefile, "/yPosStack 8 array def\n");
  fprintf (thefile, "/PosStackPos 0 def\n");

  if (dsc) fprintf (thefile, "%%%%EndSetup\n");
}

/*
 * finalize ps device
 */

void
ps_finish (FILE *thefile)
{
  if (dsc) fprintf (thefile, "%%%%Trailer\n");
  fprintf (thefile, "%%%%EOF\n");
}

/*
 * init kyo defaults
 */

void
kyo_setup (FILE *thefile)
{
  curpsize.xpos = psizes[PS_DEFAULT].xpos;
  curpsize.ypos = psizes[PS_DEFAULT].ypos;

  sbm = curpsize.ypos - unprintable.ypos;
  srm = curpsize.xpos - unprintable.xpos;

  unit      = UNIT_INCH;
  umult     = umumap[unit];
  page      = 1;
  emptypage = 1;
}

/*
 * Dummy implementation of yywrap(); we don't need to wrap files
 */

int
yywrap(void)
{
  return 1;
}

/*
 * Parse a CPP command line into the different argv's
 */

int
parsecmd (char *argv[], char *cmd)
{
  int count=0;

  argv[0] = strtok (cmd, " ");

  while (argv[count] && count<14) {
    argv[++count] = strtok (NULL, " ");
  }

  /*
   * Add `-P' option
   */

  argv[count++] = "-P";
  argv[count++] = "-C";
  argv[count]   = NULL;

  return count;
}

extern FILE * yyin;

int
main (int argc, char *argv[])
{
  int domacro=0, doall=0, status;
#ifdef HAVE_PIPE
  char *cppcom[16], rawcpp[512]=CPP;
  pid_t mppid, cppid, pid;
  int pipes[2][2];
#endif
  char theoutname[512];
  char *argv0;

  outfile   = stdout;
  lineno    = 1;
  kyoout    = 0;
  yydebug   = 0;
  dsc       = 1;
  encoding  = ISO_LATIN_1;
  outfname[0] = '\0';

  if ((inputfiles = (char **) malloc ((argc+1)*sizeof(char*))) == NULL) {
    fprintf (stderr, "oops: cannot allocate input file structure\n");
    exit (1);
  }

  inputfile=0;

  argc--; argv0=*argv++;
  while (argc) {
    if (**argv == '-' && *(*argv+1)) {
      switch (*++*argv) {
      case 'd': yydebug = 1; break;
      case 'c': kyoout  = 1; break;
      case 'm': domacro++;   break;
      case 'a': doall   = 1; break;
      case 'p': dsc     = 0; break;
      case 'f': if (argc>1) { ftmd = atoi (*++argv); argc--; } break;
      case 'u':
	switch (*(*argv+1)) {
	case 'x': if (argc>1) {unprintable.xpos=atof(*++argv);argc--;} break;
	case 'y': if (argc>1) {unprintable.ypos=atof(*++argv);argc--;} break;
	default:  if (argc>1) {unprintable.xpos=
			       unprintable.ypos=atof(*++argv);argc--;} break;
	}
	break;
      case 't':
	switch (*(*argv+1)) {
	case 'x': if (argc>1) {offset.xpos=atof(*++argv);argc--;} break;
	case 'y': if (argc>1) {offset.ypos=atof(*++argv);argc--;} break;
	default:  if (argc>1) {offset.xpos=
			       offset.ypos=atof(*++argv);argc--;} break;
	}
	break;
      case 'x': if (argc>1) { unprintable.xpos = atof (*++argv); argc--; } break;
      case 'y': if (argc>1) { unprintable.ypos = atof (*++argv); argc--; } break;
      case 's': if (argc>1) { scale = atof (*++argv); argc--; } break;
      case 'e':
	if (argc>1) {
	  argc--; argv++;
	  switch (**argv) {
	  case 'I': case 'i': encoding = ISO_LATIN_1; break;
	  case 'L': case 'l': encoding = LASERJET;    break;
	  case 'P': case 'p': encoding = IBMPC;       break;
	  default:  fprintf (stderr, "error: unknown encoding: `%s'\n", *argv);
	            exit (1);
	  }
	}
	break;
      case 'o': if (argc>1) { strcpy (outfname, *++argv); argc--; } break;
      default:
	printf ("\
usage: %s [Options ...] [file ...]\n\
\n\
  Convert Kyocera Prescribe to PostScript\n\
\n\
  -f ftmd   Set initial font mode (see ftmd)\n\
  -c        Include Prescribe commands in Postscript output as comments\n\
  -m        Don't do Postscript conversion, but perform mcro processing\n%s\
  -p        Don't print Postscript DSC comments\n\
  -d        Print scanner & parser debugging information (on stderr)\n\
  -s scale  Scale the output (>1: enlarge, <1: shrink)\n\
  -tx/ty of Translate right or up\n\
  -ux/uy mg Adjust left/top unprintable area\n\
  -e enc    Select encoding (iso8859-1, laserjet, pc)\n\
  -o name   Write to output file (can contain %%d for page no)\n\
\n\
",
		argv0,
#ifdef HAVE_PIPE
		"  -a        Do mcro processing, mcro replacing (using cpp) and conversion\n"
#else
		""
#endif
		);
	exit (1);
      }
    } else {
      inputfiles[inputfile++] = *argv;
    }
    argv++, argc--;
  }

  if (inputfile == 0) {
    inputfiles[inputfile++] = "-";
  }

  inputfiles[inputfile] = NULL;
  inputfile = 0;

  if (inputfiles[0][0]!='-' || inputfiles[0][1]!='\0') {
    if ((yyin = fopen (inputfiles[inputfile], "r")) == NULL) {
      perror (inputfiles[inputfile]);
      exit (1);
    }
  } else {
    yyin = stdin;
  }

  if (*outfname) {
    sprintf (theoutname, outfname, 1);
    if ((outfile = fopen (theoutname, "w")) == NULL) {
      fprintf (stderr, "error: cannot open output file `%s' for writing.\n",
	       theoutname);
      exit (1);
    }
  }

  if (domacro == 1) { /* only macro resolution */
    mcro ();
  }
#ifdef HAVE_PIPE
  else if (doall) {
    /*
     * need a couple of pipes for our three processes
     *  1) macro replacement   -->  C Preprocessor
     *  2) C Preprocessor      -->  P2P Conversion
     */
    if (pipe (pipes[0]) != 0) {
      perror ("couldn't create first set of pipes");
      exit (1);
    }
    if (pipe (pipes[1]) != 0) {
      perror ("couldn't create second set of pipes");
      exit (1);
    }
    if ((mppid = fork ()) == 0) {
      /*
       * first child. we are supposed to read a source file from stdin
       * and process the kyocera macros
       */
      close (pipes[0][0]);    /* close other pipes           */
      close (pipes[1][0]);    /* close other pipes           */
      close (pipes[1][1]);    /* close other pipes           */
      dup2  (pipes[0][1], 1); /* replace stdout with pipe    */
      mcro  ();               /* macro processing            */
      exit  (0);              /* finished                    */
    }
    else if (mppid == (pid_t) -1) {
      perror ("couldn't fork once");
      exit (1);
    }
    if ((cppid = fork ()) == 0) {
      /*
       * second child. we are running the C preprocessor
       */
      close (pipes[0][1]);    /* close other pipes           */
      close (pipes[1][0]);    /* close other pipes           */
      dup2  (pipes[0][0], 0); /* stream from macro processor */
      dup2  (pipes[1][1], 1); /* going to the p2p converter  */
      parsecmd (cppcom, rawcpp);
      execvp (cppcom[0], cppcom); /* exec preprocessor */

      /*
       * only executed if execl fails
       */
      perror ("couldn't exec " CPP);
      exit (1);
    }
    else if (cppid == (pid_t) -1) {
      perror ("couldn't fork twice");
      kill  (mppid, SIGINT);
      while (mppid)
	if (waitpid (mppid, &status, 0) == mppid)
	  if (WIFEXITED(status))
	    mppid = 0;
      exit  (1);
    }
    /*
     * our task is to p2p-convert the output of the preprocessor
     */
    close (pipes[0][0]);      /* close other pipes            */
    close (pipes[0][1]);      /* close other pipes            */
    close (pipes[1][1]);      /* close other pipes            */
    dup2  (pipes[1][0], 0);   /* input from preprocessor pipe */

    ps_setup  (outfile);

    /*
     * Don't use files from the command line, but stdin
     */

    yyin = stdin;
    inputfile = 0;
    inputfiles[0] = "-";
    inputfiles[1] = NULL;

    kyo_setup (outfile);
    yyparse   ();
    ps_finish (outfile);

    /*
     * if we aborted with an error, but our children are still alive,
     * this should send them a SIGPIPE
     */

    close (0);

    /*
     * wait for our children to finish
     */
    while (mppid || cppid) {
      pid = waitpid ((pid_t)-1, &status, 0);
      if (WIFEXITED(status)) {
	if (pid == mppid) mppid = 0;
	if (pid == cppid) cppid = 0;
      }
    }
  }
#else
  else if (doall) {
    fprintf (stderr, "Cannot fork preprocessor, using -a not possible\n");
  }
#endif
  else {
    ps_setup  (outfile);
    kyo_setup (outfile);
    yyparse   ();
    ps_finish (outfile);
  }

  if (outfile && outfile != stdout) {
    fclose (outfile);
  }

  return 0;
}
