/*
 * This file is part of the Prescribe to Postscript converter by
 * Frank Pilhofer <fp@informatik.uni-frankfurt.de>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * For non-commercial use ONLY.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include "p2p.h"
#include "p2p.tab.h"

char * p2p_mcro = "$Id: mcro.c,v 1.6 2000/04/05 14:13:09 fp Exp $";

#define MAXMACROPARMS	16	/* maximum number of parameters  */
#define MAXMPLENGTH	128	/* maximum length of a parameter */

static char  parbuf[MAXMACROPARMS+1][MAXMPLENGTH];
extern char *yytext;
extern int   retws;

/*
 * This program reads a Kyocera Printer File and replaces all
 * macro definitions and calls by cpp macros, so that they can
 * be resolved by /lib/cpp. Reads input from stdin and writes
 * to stdout
 */

void
call_mcro (int oc)
{
  int result, rept=1, index=0, last=0, i;
  char *p1, *p2, temp[256];

  if ((result = yylex()) == 0)
    return;
  while (result == WS)
    if ((result = yylex()) == 0)
      return;

  if (result == NUMBER) {
    rept   = (int) (yylval.v_number+0.5);
    result = yylex();
  }

  /*
   * get name of macro
   */

  while (result == WS)
    if ((result = yylex ()) == 0)
      return;

  if (result == OPTIONS) {
    sprintf (parbuf[0], "mcro_");
    for (i=0; i<4 && yytext[i]; i++)
      parbuf[0][5+i] = tolower(yytext[i]);
    parbuf[0][5+i] = '\0';
  }
  else if (result == oc) {
    if ((result = yylex()) == 0)
      return;
    if (result == NUMBER) {
      sprintf (parbuf[0], "_mac_par_%03d",
	       (int) (yylval.v_number+.5));
    }
  }

  /*
   * process parameter
   */

  while (42) {
    last = result;
    if ((result = yylex()) == 0)
      break;
    if (result == ';')
      break;
    if (result == ',') {
      parbuf[++index][0] = '\0';
      continue;
    }
    else if (result == WS) {
      if (last == ',') {
	result = last;
	continue;
      }
      strcat (parbuf[index], yytext);
      continue;
    }
    else if (result == oc) {
      if ((result = yylex()) == 0)
	return;
      if (result == NUMBER) {
	sprintf (temp, "_mac_par_%03d",
		 (int) (yylval.v_number+.5));
	strcat  (parbuf[index], temp);
      }
    }
    else if (result == STRING) {
      if (strchr (yylval.v_string, oc) == NULL)
	strcat (parbuf[index], yytext);
      else {
	p1 = parbuf[index] + strlen(parbuf[index]);
	p2 = yylval.v_string;

	*p1++ = yytext[0];
	while (*p2) {
	  if (*p2 == oc) {
	    sprintf (p1, "_mac_par_%03d",
		     (int) strtol (++p2, &p2, 0));
	    p1 += 12;
	  }
	  else {
	    *p1++ = *p2++;
	  }
	}
	*p1++ = yytext[0];
	*p1   = '\0';
      }
    }
    else {
      strcat (parbuf[index], yytext);
    }
  }

  /*
   * nullify other parameters
   */

  for (index++; index<MAXMACROPARMS; index++)
    parbuf[index][0] = '\0';

  /*
   * print macro calls
   */

  for ( ; rept; rept--) {
    fprintf (stdout, " %s(",parbuf[0]);
    for (index=1; index<=MAXMACROPARMS; index++)
      fprintf (stdout, "%s%c",
	       parbuf[index],
	       (index==MAXMACROPARMS)?')':',');
  }

  fputc (' ', stdout);
  return;
}

void
define_mcro ()
{
  int result, thechar='%', i, last;
  char *p;

  if ((result = yylex()) == 0)
    return;
  while (result == WS)
    if ((result = yylex()) == 0)
      return;

  if (result != OPTIONS) {
    fprintf (stdout, "mcro %s", yytext);
    return;
  }
  fprintf (stdout, "\n#define mcro_");

  for (i=0; i<4 && yytext[i]; i++)
    fputc (tolower(yytext[i]), stdout);

  fputc ('(', stdout);
  for (i=1; i<=MAXMACROPARMS; i++)
    fprintf (stdout, "_mac_par_%03d%c", i,
	     (i==MAXMACROPARMS)?')':',');
  fputc (' ', stdout);

  if ((result = yylex()) == 0)
    return;
  while (result == WS)
    if ((result = yylex()) == 0)
      return;

  if (result == ',') {
    if ((result = yylex()) == 0)
      return;
    while (result == WS)
      if ((result = yylex()) == 0)
	return;

    thechar = (yytext[0]==';')?'%':yytext[0];
  }
  else if (result == 0)
    return;

  /*
   * ignore remainder of line
   */

  while (result != ';')
    if ((result = yylex()) == 0)
      break;

  /*
   * macro begins here
   */

  while (42) {
    if ((result = yylex()) == 0)
      return;
    if (result == ENDM)
      break;
    else if (result == thechar) {
      if ((result = yylex()) == 0)
	return;
      if (result == NUMBER) {
	fprintf (stdout, "_mac_par_%03d",
		 (int) (yylval.v_number+.5));
      }
      else {
	fprintf (stdout, "_mac_par_001 %s", yytext);
      }
    }
    else {
      switch (result) {
      case '\n':
      case '\r':
	fprintf (stdout, " \\\n\t");
	break;
      case '\f':
	/* do nothing */
	break;
      case CALL:
	call_mcro (thechar);
	break;
      case OPTIONS:
	if (strchr (yylval.v_string, '/') == NULL) {
	  fprintf (stdout, "%s", yytext);
	}
	else {
	  p = yylval.v_string;
	  last = 0;
	  while (*p) {
	    if (*p == '/' && last == '/') {
	      fputc (' ', stdout);
#if 0
	      fputc ('#', stdout);
	      fputc ('#', stdout);
	      fputc (' ', stdout);
#endif
	    }
	    fputc (*p, stdout);
	    last = *p++;
	  }
	}
	break;
      case STRING:
	if (strchr (yylval.v_string, thechar) == NULL) {
	  fprintf (stdout, "%s", yytext);
	}
	else {
	  fputc (yytext[0], stdout);
	  p = yylval.v_string;
	  while (*p) {
	    if (*p == thechar) {
	      fprintf (stdout, "_mac_par_%03d",
		       (int) strtol (++p, &p, 0));
	    }
	    else
	      fputc (*p++, stdout);
	  }
	  fputc (yytext[0], stdout);
	}
	break;
      default:
	fprintf (stdout, "%s", yytext);
	break;
      }
    }
  } /* while (42) */

  /*
   * end of macro definition, skip eol
   */
  while (result!=0 && result!=';')
    result = yylex();

  fprintf (stdout, "\n");
}

void
undef_mcro ()
{
  int result, i;

  if ((result = yylex()) == 0)
    return;
  while (result == WS)
    if ((result = yylex()) == 0)
      return;

  if (result != OPTIONS) {
    fprintf (stdout, "delm %s", yytext);
    return;
  }
  fprintf (stdout, "\n#ifdef mcro_");
  for (i=0; i<4 && yytext[i]; i++)
    fputc (tolower(yytext[i]), stdout);
  fprintf (stdout, "\n#undef mcro_");
  for (i=0; i<4 && yytext[i]; i++)
    fputc (tolower(yytext[i]), stdout);
  fprintf (stdout, "\n#endif\n");

  result = yylex (); /* munch semicolon */
  return;
}

void
mcro ()
{
  int result;

  retws = 1;

  while (42) {
    if ((result = yylex()) == 0)
      break;
    switch (result) {
    case MCRO: 
      define_mcro ();
      break;
    case CALL:
      call_mcro (0);
      break;
    case DELM:
      undef_mcro ();
      break;
    default:
      fprintf (stdout, "%s", yytext);
      break;
    }
  }
  fprintf (stdout, "\n");
}
