/*
 * $Id: p2p.h,v 1.13 2000/06/25 10:55:18 fp Exp $
 */

/*
 * Druckerstatus
 */

#define UNIT_INCH	0	/* Unit measurements */
#define UNIT_CM		1
#define UNIT_DOT	2
#define UNIT_PIXEL	3

#define CASS_MANUAL	0	/* Paper feed */
#define CASS_UPPER	1
#define CASS_LOWER	2

/*
 * Default Paper Size
 */

#define PS_DEFAULT	8

#define PS_CASS		0	/* Paper sizes */
#define PS_MONARCH	1	/* 3 7/8 x 7  1/2 in */
#define PS_BUSINESS	2	/* 4 1/8 x 9  1/2 in */
#define PS_INTDL	3	/* 11    x 22     cm */
#define PS_INTC5	4	/* 16.2  x 22.9   cm */
#define PS_EXECUTIVE	5	/* 7 1/4 x 10 1/2 in */
#define PS_LETTER	6	/* 8 1/2 x 11     in */
#define PS_LEGAL	7	/* 8 1/2 x 14     in */
#define PS_A4		8	/* 21    x 29.7   cm */
#define PS_B5		9	/* 18.2  x 25.7   cm */

#define PO_PORTRAIT	0	/* Paper orientation */
#define PO_LANDSCAPE	1

/*
 * Max no of fonts
 */

#define MAXFONT		79

/*
 * (x,y) structure
 */

typedef struct {
  double xpos, ypos;
} point;

/*
 * Font mapping structure
 */

typedef struct {
  char    font[64];		/* postscript font name       */
  char    sfnt[64];		/* sfnt font name             */
  double  size;         	/* font size                  */
  double  cs;           	/* character width, 0 if prop */
  int     po;			/* 0: portrait, 1: landscape  */
} fm;

/*
 * Structure to hold LDFC'd characters
 */

typedef struct _ldfc {
  int fontnum, code;		/* font number and ASCII code */
  int height, width;		/* pixel                      */
  int yoffset, xoffset;		/* micropixel (1/32 pixel)    */
  int cellwidth, cellctr;	/* micropixel (1/32 pixel)    */
  int rotation;			/* *90 deg right              */
  unsigned char *data;		/* pixel data                 */
  int xi, yi;			/* pointer while loading      */
  struct _ldfc *next;
} ldfc;

/*
 * structure for font information (LDFC, SFA, GENF)
 */

typedef struct _ldfont {
  int page;			/* last used on which page    */
  int fontnum;			/* font number                */
  enum {
    LDFC_FONT,                  /* bitmap font using ldfc     */
    GENF_FONT,                  /* dynamic font using genf    */
    SFNT_FONT                   /* typeface set with sfnt     */
  } type;
  struct {
    double minx, maxx;
    double miny, maxy;          /* bounding box for font      */
    int rotation;               /* *90 deg right              */
    ldfc * chars;               /* character data             */
  } ldfc;
  struct {
    int sfa, fp;                /* if set with sfa            */
    double hmi, vmi;            /* if set with sfa            */
  } sfa;
  struct {
    int genf;
    char name[64];              /* dynamic font name          */
    double height;
    int first, last, target;
    double comp, ital;
    int rot, type;
    int xbold, ybold;
  } genf;
  struct {
    int base;                   /* prescribe font number      */
    char name[64];              /* postscript font name       */
    double height;
    int symbol_set;
    double compression;
    double shear;
  } sfnt;
  struct _ldfont *next;
} ldfont;

/*
 * Structure for RVRD raster data, each node holds a single line
 */

typedef struct _rvrd {
  int length;			/* this line's length         */
  unsigned char *data;		/* this line's data           */
  int xi;			/* pointer while loading      */
  struct _rvrd *next;		/* pointer to next line       */
} rvrd;

enum encoding {
  ISO_LATIN_1,
  LASERJET,
  IBMPC
};

extern double stm, slm, srm, sbm, spd;
extern double sls, scs, sulpd, sulps;
extern int font, ftmd, prop, spo, str;
extern point curpos, cpstack[];
extern int cpspos, unit, fontsp;
extern const fm fontmap[], scalable[];
extern int altf[], fonts[];
extern unsigned char fpat[];
extern unsigned char prepats[][32];
extern unsigned char xpats[][32];
extern point curpsize;
extern int page, emptypage;
extern FILE *outfile;
extern ldfont *ldfonts;
extern ldfont curfont;
extern enum encoding encoding;

extern double umult, scale;
extern int dsc, kyoout, lineno;
extern const double pixel;
extern const point psizes[];
extern const double umumap[];
extern point unprintable;
extern point offset;

extern char **inputfiles;
extern int inputfile;
extern char outfname[];

void    ps_setup        (FILE *);
void    ps_finish       (FILE *);

ldfont *havefont	(int);
void	kyoldfc		(ldfc *);
void	kyodelf		(int);
void	kyodaf		(void);
void	kyosfa		(int, double, double, int);
void    kyosfnt         (char *, double, int, int, double, double);
void	loadfont	(int);
void	dumpfonts	(FILE *);
void	kyorvrd		(rvrd *);
void	kyokillrd	(rvrd *);
void	kyofpat		(unsigned char *);
void	kyofont		(void);
void	kyofontprops	(int, double, double, int, int, int);
void	initpage	(void);
void	endpage		(void);
char *	outtext		(char *);
void    mcro            ();

