%{

/*
 * This file is part of the Prescribe to Postscript converter by
 * Frank Pilhofer <fp@informatik.uni-frankfurt.de>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * For non-commercial use ONLY.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "p2p.h"

char * p2p_p2p_y = "$Id: p2p.y,v 1.17 2000/09/19 09:38:34 fp Exp $";

/*
 * DOS doesn't have this. Grr.
 */

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define YYDEBUG		1
#define YYERROR_VERBOSE

/*
 * constants
 */

/*
 * unit multiplier to get to points
 */

const double umumap[4] = {
  72.0, 28.3464567, 1.0, 0.24
};

const double pixel = 0.24; /* umumap[UNIT_PIXEL] */

/*
 * scaling factor
 */

double scale = 1.0;

/*
 * Paper sizes in points (1in=72 points)
 */

const point psizes[10] = {
  {  595.2756,  841.8898 }, /* default size, A4 */
  {  279.0000,  540.0000 }, /* MONARCH          */ /* 3 7/8 x 7  1/2 in */
  {  297.0000,  684.0000 }, /* BUSINESS         */ /* 4 1/8 x 9  1/2 in */
  {  311.8110,  623.6220 }, /* INTDL            */ /* 11    x 22     cm */
  {  459.2126,  649.1339 }, /* INTC5            */ /* 16.2  x 22.9   cm */
  {  522.0000,  756.0000 }, /* EXECUTIVE        */ /* 7 1/4 x 10 1/2 in */
  {  612.0000,  792.0000 }, /* LETTER           */ /* 8 1/2 x 11     in */
  {  612.0000, 1008.0000 }, /* LEGAL            */ /* 8 1/2 x 14     in */
  {  595.2756,  841.8898 }, /* A4               */ /* 21    x 29.7   cm */
  {  515.9055,  728.5039 }  /* B5               */ /* 18.2  x 25.7   cm */
};

/*
 * unprintable margin in x and y direction, (3mm,7mm)
 */

/* point unprintable = { 18.5, 23.5 }; */ /* { 8.5, 13.5 }; */
point unprintable = { 13.5, 13.5 }; /* { 8.5, 13.5 }; */

/*
 * offset, for printing to other printers
 */

point offset = { 0.0, 0.0 };

/*
 * (default) spacings and margins, in points where appropriate
 */

double stm   = 0.0;	/* top margin */
double slm   = 0.0;	/* left margin */
double srm   = 0.0;	/* right margin */
double sbm   = 0.0;	/* bottom margin */
double sls   = 12.0;	/* line spacing, 6 lpi */
double scs   = 7.2;	/* character spacing, 10cpi */
double sulpd = 1.68;	/* underline distance, 7 pixel */
double sulps = 0.96;	/* underline strength, 4 pixel */

/*
 * other (default) parameters
 */

double spd   = 0.72;	/* pen diameter 3 pixel */
int font     = 1;	/* font */
int ftmd     = 15;	/* font mode */
int prop     = 0;	/* proportional font */
int spo      = 0;	/* 0: portrait, 1: landscape, -1 for initialization */
int copies   = 1;	/* number of copies, currently ignored */
int sro      = 0;	/* raster image orientation */

/*
 * cursor position and cursor position stack
 */

point curpos;		/* xpos may be wrong if using proportional fonts */

point cpstack[8];
int cpspos=0;

/*
 * font map prescribe -> postscript
 *
 *  postscript name     prescribe sfnt name         size/width/po
 */

const fm fontmap[80] = {
  { "Illegal-Font",     "",                         12.0, 6.00, 0 },
  { "Courier",          "CourierBM12-Roman",        12.0, 7.20, 0 }, /* 01 */
  { "Times-Roman",      "Dutch801BM10-Roman",       10.0, 0.00, 0 }, /* 02 */
  { "Times-Italic",     "Dutch801BM10-Italic",      10.0, 0.00, 0 }, /* 03 */
  { "Times-Bold",       "Dutch801BM10-Bold",        10.0, 0.00, 0 }, /* 04 */
  { "Times-Roman",      "Dutch801BM8-Roman",         8.0, 0.00, 0 }, /* 05 */
  { "Courier",          "PrestigeEliteBM10-Roman",  10.0, 6.00, 0 }, /* 06 */
  { "Courier",          "PrestigeEliteBM7.2-Roman",  7.2, 4.33, 0 }, /* 07 */
  { "Courier",          "LetterGothicBM12-Roman",   12.0, 6.00, 0 }, /* 08 */
  { "Courier-Bold",     "LetterGothicBM12-Bold",    12.0, 6.00, 0 }, /* 09 */
  { "Helvetica-Bold",   "Swiss721BM14.4-Bold",      14.4, 0.00, 0 }, /* 10 */
  { "Helvetica-Bold",   "Swiss721BM12-Bold",        12.0, 0.00, 0 }, /* 11 */
  { "Helvetica-Bold",   "Swiss721BM10-Bold",        10.0, 0.00, 0 }, /* 12 */
  { "Helvetica",        "Swiss721BM8-Roman",         8.0, 0.00, 0 }, /* 13 */
  { "Helvetica",        "Swiss721BM6-Roman",         6.0, 0.00, 0 }, /* 14 */
  { "Courier",          "LinePrinterBM9-Roman",      9.0, 4.33, 0 }, /* 15 */
  { "Courier",          "LinePrinterBM7-Roman",      6.0, 3.36, 0 }, /* 16 */
/*					    
 * Some Landscape Fonts			    
 */					    
  { "Courier",          "CourierBM12-Roman",        12.0, 7.20, 1 }, /* 17 */
  { "Courier-Italic",   "CourierBM12-Italic",       12.0, 7.20, 1 }, /* 18 */
  { "Times-Roman",      "Dutch801BM10-Roman",       10.0, 0.00, 1 }, /* 19 */
  { "Times-Italic",     "Dutch801BM10-Italic",      10.0, 0.00, 1 }, /* 20 */
  { "Times-Bold",       "Dutch801BM10-Bold",        10.0, 0.00, 1 }, /* 21 */
  { "Times-Roman",      "Dutch801BM8-Roman",         8.0, 0.00, 1 }, /* 22 */
  { "Courier",          "PrestigeEliteBM10-Roman",  10.0, 6.00, 1 }, /* 23 */
  { "Courier-Italic",   "PrestigeEliteBM10-Italic", 10.0, 6.00, 1 }, /* 24 */
  { "Courier",          "PrestigeEliteBM7.2-Roman",  7.2, 4.33, 1 }, /* 25 */
  { "Courier",          "LetterGothicBM12-Roman",   12.0, 6.00, 1 }, /* 26 */
  { "Courier-Italic",   "LetterGothicBM12-Italic",  12.0, 6.00, 1 }, /* 27 */
  { "Courier-Bold",     "LetterGothicBM12-Bold",    12.0, 6.00, 1 }, /* 28 */
  { "Helvetica-Bold",   "Swiss721BM14.4-Bold",      14.4, 0.00, 1 }, /* 29 */
  { "Helvetica-Bold",   "Swiss721BM12-Bold",        12.0, 0.00, 1 }, /* 30 */
  { "Helvetica-Bold",   "Swiss721BM10-Bold",        10.0, 0.00, 1 }, /* 31 */
  { "Helvetica",        "Swiss721BM8-Roman",         8.0, 0.00, 1 }, /* 32 */
  { "Helvetica",        "Swiss721BM6-Roman",         5.7, 0.00, 1 }, /* 33 */
  { "Courier",          "LinePrinterBM9-Roman",      9.0, 4.33, 1 }, /* 34 */
  { "Courier-Italic",   "LinePrinterBM9-Italic",     9.0, 4.33, 1 }, /* 35 */
  { "Courier",          "LinePrinterBM7-Roman",      6.0, 3.36, 1 }, /* 36 */
/*					    
 * extended portrait fonts		    
 */					    
  { "Courier-Italic",   "CourierBM12-Italic",       12.0, 7.20, 0 }, /* 37 */
  { "Courier-Bold",     "CourierBM12-Bold",         12.0, 7.20, 0 }, /* 38 */
  { "Courier-BoldItalic","CourierBM12-BoldItalic",  12.0, 7.20, 0 }, /* 39 */
  { "Times-BoldItalic", "Dutch801BM10-BoldItalic",  10.0, 0.00, 0 }, /* 40 */
  { "Times-Italic",     "Dutch801BM8-Italic",        8.0, 0.00, 0 }, /* 41 */
  { "Times-Bold",       "Dutch801BM8-Bold",          8.0, 0.00, 0 }, /* 42 */
  { "Times-BoldItalic", "Dutch801BM8-BoldItalic",    8.0, 0.00, 0 }, /* 43 */
  { "Courier-Italic",   "PrestigeEliteBM10-Italic", 10.0, 6.00, 0 }, /* 44 */
  { "Courier-Bold",     "PrestigeEliteBM10-Bold",   10.0, 6.00, 0 }, /* 45 */
  { "Courier-BoldItalic","PrestigeEliteBM10-BoldItalic",10.0,6.00,0},/* 46 */
  { "Courier-Italic",   "PrestigeEliteBM7.2-Italic", 7.2, 4.33, 0 }, /* 47 */
  { "Courier-Bold",     "PrestigeEliteBM7.2-Bold",   7.2, 4.33, 0 }, /* 48 */
  { "Courier-BoldItalic","PrestigeEliteBM7.2-BoldItalic",7.2,4.33,0},/* 49 */
  { "Courier-Italic",   "LetterGothicBM12-Italic",  12.0, 6.00, 0 }, /* 50 */
  { "Courier-BoldItalic","LetterGothicBM12-BoldItalic",12.0,6.00,0}, /* 51 */
  { "Helvetica-Oblique","Swiss721BM14.4-BoldItalic",14.4, 0.00, 0 }, /* 52 */
  { "Helvetica-BoldOblique","Swiss721BM12-BoldItalic",12.0,0.00,0 }, /* 53 */
  { "Helvetica-BoldOblique","Swiss721BM10-BoldItalic",10.0,0.00,0 }, /* 54 */
  { "Helvetica-Oblique","Swiss721BM8-Italic",        8.0, 0.00, 0 }, /* 55 */
  { "Helvetica-Bold",   "Swiss721BM8-Bold",          8.0, 0.00, 0 }, /* 56 */
  { "Helvetica-BoldOblique","Swiss721BM8-BoldItalic",8.0, 0.00, 0 }, /* 57 */
  { "Helvetica-Oblique","Swiss721BM6-Italic",        6.0, 0.00, 0 }, /* 58 */
  { "Helvetica-Bold",   "Swiss721BM6-Bold",          6.0, 0.00, 0 }, /* 59 */
  { "Helvetica-BoldOblique","Swiss721BM6-BoldItalic",6.0, 0.00, 0 }, /* 60 */
  { "Courier-Italic",   "LinePrinterBM9-Italic",     9.0, 4.33, 0 }, /* 61 */
  { "Courier-Bold",     "LinePrinterBM9-Bold",       9.0, 4.33, 0 }, /* 62 */
  { "Courier-BoldItalic","LinePrinterBM9-BoldItalic",9.0, 4.33, 0 }, /* 63 */
  { "Courier-Italic",   "LinePrinterBM7-Italic",     6.0, 3.36, 0 }, /* 64 */
  { "Courier-Bold",     "LinePrinterBM7-Bold",       6.0, 3.36, 0 }, /* 65 */
  { "Courier-BoldItalic","LinePrinterBM7-BoldItalic",6.0, 3.36, 0 }, /* 66 */
/*
 * extended landscape fonts
 */
  { "Courier-Bold",     "CourierBM12-Bold",         12.0, 7.20, 1 }, /* 67 */
  { "Courier-BoldItalic","CourierBM12-BoldItalic",  12.0, 7.20, 1 }, /* 68 */
  { "Times-BoldItalic", "Dutch801BM10-BoldItalic",  10.0, 0.00, 1 }, /* 69 */
  { "Times-Bold",       "Dutch801BM8-Bold",          8.0, 0.00, 1 }, /* 70 */
  { "Courier-Bold",     "PrestigeEliteBM10-Bold",   10.0, 6.00, 1 }, /* 71 */
  { "Courier-BoldItalic","PrestigeEliteBM10-BoldItalic",10.0,6.00,1},/* 72 */
  { "Courier-Bold",     "PrestigeEliteBM7.2-Bold",   7.2, 4.33, 1 }, /* 73 */
  { "Helvetica-Bold",   "Swiss721BM8-Bold",          8.0, 0.00, 1 }, /* 74 */
  { "Helvetica-Bold",   "Swiss721BM6-Bold",          6.0, 0.00, 1 }, /* 75 */
  { "Courier-Bold",     "LinePrinterBM9-Bold",       9.0, 4.33, 1 }, /* 76 */
  { "Courier-BoldItalic","LinePrinterBM9-BoldItalic",9.0, 4.33, 1 }, /* 77 */
  { "Courier-Bold",     "LinePrinterBM7-Bold",       6.0, 3.36, 1 }, /* 78 */
  { "Courier-BoldItalic","LetterGothicBM12-BoldItalic",12.0,6.00,1}, /* 79 */
};

/*
 * Scalable fonts for use with sfnt
 */

const fm scalable[] = {
  { "Times-Bold",       "Dutch801SWC-Bold",          0.0, 0.00, 0 },
  { "Times-BoldItalic", "Dutch801SWC-BoldItalic",    0.0, 0.00, 0 },
  { "Times-Italic",     "Dutch801SWC-Italic",        0.0, 0.00, 0 },
  { "Times-Roman",      "Dutch801SWC-Roman",         0.0, 0.00, 0 },
  { "Helvetica-Bold",   "Swiss742SWC-Bold",          0.0, 0.00, 0 },
  { "Helvetica-Bold",   "Swiss742SWC-BoldCondensed", 0.0, 0.00, 0 },
  { "Helvetica-BoldOblique","Swiss742SWC-BoldCondensedItalic",0.0,0.00,0},
  { "Helvetica-BoldOblique","Swiss742SWC-BoldItalic",0.0, 0.00, 0 },
  { "Helvetica-Oblique","Swiss742SWC-Italic",        0.0, 0.00, 0 },
  { "Helvetica-Oblique","Swiss742SWC-ItalicCondensed",0.0,0.00, 0 },
  { "Helvetica",        "Swiss742SWC-Roman",         0.0, 0.00, 0 },
  { "Helvetica",        "Swiss742SWC-RomanCondensed",0.0, 0.00, 0 },
  { "",                 "",                          0.0, 0.00, 0 }
};



/*
 * Runtime Parameters
 */

/*
 * unit and current multiplier
 */

int unit = UNIT_INCH;
double umult;

/*
 * alternative fonts and font stack
 */

int altf[32];
int fonts[8], fontsp=0;

/*
 * fill pattern, 16x16 bit mask
 */

unsigned char fpat[32]  = {
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255
};

/*
 * extended fill patterns
 */

int xpdef, xpdi;
unsigned char xpats[6][32];

/*
 * pie pieces
 */

int piepieces;
double pieslices[128];

/*
 * current paper size
 */

point curpsize;

/*
 * current page
 */

int page = 0;

/*
 * is page empty yet?
 */

int emptypage = 1;

/* 
 * output file
 */

FILE *outfile;
char outfname[512];

/*
 * linked list for loaded and dynamic fonts
 */

ldfont * ldfonts = NULL;

/*
 * current font, if font==0
 */

ldfont curfont;

/*
 * structure for newly loaded character
 */

ldfc * newchar;

/*
 * raster image resolution
 */

int str = 75;

/*
 * currently-loading raster data line
 */

rvrd * newrd;

/*
 * encoding
 */

enum encoding encoding;

/*
 * setup kyo font
 */

void
kyofont (void)
{
  ldfont *tf;

  if (font == 0) {
    tf = &curfont;
  }
  else {
    tf = havefont (font);
  }

  if (font == 0 || font > 1000) {
    if (!emptypage) {
      loadfont (font);
    }
    if (tf && tf->sfa.sfa) {
      kyofontprops (ftmd,
		    tf->sfa.vmi,
		    tf->sfa.hmi,
		    tf->sfa.fp,
		    spo, !emptypage);
    }
    else if (tf && tf->type == SFNT_FONT) {
      if (tf->sfnt.base) {
	kyofontprops (ftmd,
		      fontmap[tf->sfnt.base].size,
		      fontmap[tf->sfnt.base].cs,
		      (fontmap[tf->sfnt.base].cs <= 0.0) ? 1 : 0,
		      fontmap[tf->sfnt.base].po,
		      !emptypage);
      }
      else {
	kyofontprops (ftmd, tf->sfnt.height, 0, 1, spo, !emptypage);
      }
    }
    else {
      kyofontprops (ftmd, 0.0, -1.0, 1, spo, !emptypage);
    }
  }
  else {
    if (tf && tf->sfa.sfa) {
      kyofontprops (ftmd,
		    tf->sfa.vmi,
		    tf->sfa.hmi,
		    tf->sfa.fp,
		    fontmap[font].po,
		    !emptypage);
    }
    else {
      kyofontprops (ftmd,
		    fontmap[font].size,
		    fontmap[font].cs,
		    (fontmap[font].cs <= 0.0) ? 1 : 0,
		    fontmap[font].po,
		    !emptypage);
    }
    if (!emptypage) {
      fprintf (outfile, "%lf %lf %d 0 1 %d %lf /%s-ISOLatin1 /%s kyofont\n",
	       sls * scale,
	       scs * scale,
	       prop,
	       spo,
	       fontmap[font].size * scale,
	       fontmap[font].font,
	       fontmap[font].font);
    }
  }
}

/*
 * set page orientation
 */

void
set_spo (int newspo, int doprint, int fontsel)
{
  double temp;

  if (newspo == spo)
    return;

  /*
   * adjust paper size and margins
   */

  if (spo != -1) {
    temp = curpsize.xpos;
    curpsize.xpos = curpsize.ypos;
    curpsize.ypos = temp;

    stm = 0.0;
    sbm = curpsize.ypos - unprintable.ypos;
    slm = 0.0;
    srm = curpsize.xpos - unprintable.xpos;
  }

  if (doprint) {
    fprintf (outfile, "currentpoint grestore gsave\n");

    if (newspo == 0) { /* portrait */
      fprintf (outfile, "%f %f translate %f %f scale\n",
	       unprintable.xpos + offset.xpos,
	       (curpsize.ypos - 2*unprintable.ypos)*scale + unprintable.ypos + offset.ypos,
	       1.0, -1.0);
    }
    else { /* landscape */
      fprintf (outfile, "%f %f translate 90 rotate %f %f scale\n",
	       unprintable.xpos + offset.xpos,
	       unprintable.ypos + offset.ypos,
	       1.0, -1.0);
    }
  }

  /*
   * switch to appropriate font if necessary
   */

  if (newspo == 0) {
    if (fontmap[font].po == 1) {
      font = 1;
    }
  }
  else {
    if (fontmap[font].po == 0) {
      font = 17;
    }
  }

  spo = newspo;

  if (doprint) {
    fprintf (outfile, "exch moveto\n");
    fprintf (outfile, "/spo %d def\n", spo);

    /*
     * need to restore font state
     */

    if (!fontsel) {
      kyofont ();
    }
  }
}

/*
 * set kyo font props according to current ftmd
 */

void
kyofontprops (int mode, double ls, double cs, int pr, int po, int doprint)
{
  if (mode & 8) set_spo (po, doprint, 1);
  if (mode & 4) prop = pr;
  if (mode & 2) sls = ls;
  if (mode & 1) scs = cs;
}

/*
 * initialize new page
 */

void
initpage (void)
{
  int oldspo = spo;
  char theoutname[512];
  emptypage = 1;

  if (*outfname) {
    sprintf (theoutname, outfname, page);
    if (strcmp (theoutname, outfname) != 0) {
      ps_finish (outfile);
      fclose (outfile);
      if ((outfile = fopen (theoutname, "w")) == NULL) {
	fprintf (stderr, "error: cannot open output file `%s' for writing.\n",
		 theoutname);
	exit (1);
      }
      ps_setup (outfile);
    }
  }

  if (dsc) {
    fprintf (outfile, "%%%%Page: %d %d\n", page, page);
    fprintf (outfile, "%%%%BeginPageSetup\n");
  }
  fprintf (outfile, "save gsave\n");
  fprintf (outfile, "0 0 moveto\n");
  kyofont ();
  fprintf (outfile, "\
/stm %f def /slm %f def /srm %f def /sbm %f def\n\
/sls %f def /scs %f def /sulpd %f def /sulps %f def\n\
/ftmd %d def /pixel %f def /prop %d def /str %d def\n\
/sro %d def\n\
",
	   stm*scale, slm*scale, srm*scale, sbm*scale,
	   sls*scale, scs*scale,
	   sulpd*scale, sulps*scale,
	   ftmd, pixel*scale, prop,
	   str, sro);
  
  kyofpat (fpat);
  fprintf (outfile, "%f setlinewidth\n", spd * scale);
  fprintf (outfile, "0 setlinecap\n");

  spo = -1;
  set_spo (oldspo, 1, 0);

  fprintf (outfile, "slm stm moveto\n");

  if (dsc) {
    fprintf (outfile, "%%%%EndPageSetup\n");
  }

  emptypage=0;

  kyofont ();
  fprintf (outfile, "%f %f moveto\n",
	   curpos.xpos*scale,
	   curpos.ypos*scale);
}

void
endpage (void)
{
  if (!emptypage) {
    fprintf (outfile, "showpage\n");
    if (dsc) {
      fprintf (outfile, "%%%%PageTrailer\n");
      fprintf (outfile, "grestore restore\n");
    }
    emptypage=1;
  }
  curpos.xpos = slm;
  curpos.ypos = stm;
  page++;
}

char *
outtext (char *t)
{
  static char temp[512];
  char *p = temp;
  int count=0;

  do {
    switch (*t) {
    case '\t':
      if (count%8==0) *p++=' ', count++;
      while (count%8) *p++=' ', count++;
      break;

    case '\r': /* huh? */
      *p++ = '\\'; count++;
      *p++ = 'r';  count++;
      break;

    case '\n': /* huh? */
      *p++ = '\\'; count++;
      *p++ = 'n';  count++;
      lineno++;
      break;

    case '(':
    case ')':
    case '\\':
      *p++ = '\\'; count++;
    default:
      *p++ = *t;
      count++;
    }
  }
  while (*t++);
  
  return temp;
}

void
printplaintext (char *t)
{
  static int uline = 0;
  static int bofl = 1;
  static double offset = 0;
  char *ptr, x;
  int i;

  /*
   * Handle some Laserjet Escape sequences, courtesy the examples from
   * Josef.Geissing@wintershall.de
   */

  while (*t) {
    switch (*t) {
    case '\033': /* ESC */
      if (*(t+1) == '&' && *(t+2) == 'd' && *(t+3) == 'D') {
	/* ESC&dD - underline text */
	if (kyoout) fprintf (outfile, "%% ESC&dD (underline on)\n");
	uline = 1;
	t += 4;
      }
      else if (*(t+1) == '&' && *(t+2) == 'd' && *(t+3) == '@') {
	/* ESC&d@ - don't underline text */
	if (kyoout) fprintf (outfile, "%% ESC&d@ (underline off)\n");
	uline = 0;
	t += 4;
      }
      else if (*(t+1) == '&' && *(t+2) == 'a') {
	/* ESC&a#L - set left margin */
	i = 0;
	t += 3;
	while (*t >= '0' && *t <= '9') {
	  i *= 10;
	  i += *t++ - '0';
	}
	if (*t == 'L') {
	  t++;
	}
	if (kyoout) fprintf (outfile, "%% ESC&a%dL (set left margin)\n", i);
	offset = i * scs;
      }
      else if (*(t+1) == '&' && *(t+2) == 's' &&
	       (*(t+3) == '0' || *(t+3) == '1') &&
	       *(t+4) == 'C') {
	/* ESC&s#C - end of line wrap enabled/disabled */
	if (kyoout) fprintf (outfile, "%% ESC&s?C (EOL wrap) (ignored)\n");
	t += 5;
      }
      else if (*(t+1) == 'E') {
	/* ESCE - printer reset */
	if (kyoout) fprintf (outfile, "%% ESCE (printer reset) (ignored)\n");
	t += 2;
      }
      else if (*(t+1) == '9') {
	/* ESC9 - clear horizontal margins */
	if (kyoout) fprintf (outfile, "%% ESC9 (clear horizontal margins)\n");
	offset = 0;
	t += 2;
      }
      else if (*(t+1) == '&' && *(t+2) == 'l' &&
	       *(t+3) == '0' && *(t+4) == 'O') {
	/* ESC&l0O - portrait orientation */
	if (kyoout) fprintf (outfile, "%% ESC&l0O (portrait orientation)\n");
	set_spo (0, !emptypage, 0);
	t += 5;
      }
      else if (*(t+1) == '&' && *(t+2) == 'l' &&
	       *(t+3) == '1' && *(t+4) == 'O') {
	/* ESC&l1O - landscape orientation */
	if (kyoout) fprintf (outfile, "%% ESC&l1O (landscape orientation)\n");
	set_spo (1, !emptypage, 0);
	t += 5;
      }
      else {
	if (kyoout) fprintf (outfile, "%% unknown escape sequence\n");
	t++;
      }
      break;

    case '\010': /* Backspace */
      if (kyoout) fprintf (outfile, "%% \\010 (Backspace)\n");
      if (emptypage) {
	curpos.xpos -= scs;
      }
      else {
	fprintf (outfile, "0 scs sub 0 rmoveto\n");
      }
      t++;
      break;

    case '\n':
      if (!emptypage) {
	fprintf (outfile, "5 0 () text\n");
      }
      curpos.ypos += sls;
      bofl = 1;
      t++;
      break;

    case '\r':
      if (kyoout) fprintf (outfile, "%% \\r (Carriage Return)\n");
      if (!emptypage) {
	fprintf (outfile, "currentpoint exch pop slm exch moveto\n");
      }
      curpos.xpos = slm;
      bofl = 1;
      t++;
      break;

    case '\f':
      if (!emptypage) {
	if (kyoout) fprintf (outfile, "%% \\f (Form Feed)\n");
	endpage();
      }
      else {
	if (kyoout) fprintf (outfile, "%% \\f (Form Feed) (ignored)\n");
	curpos.xpos = slm;
	curpos.ypos = stm;
      }
      t++;
      break;

      /*
       * With SFA in action, spaces can have a different width than
       * normal characters
       */

    case ' ':
      if (emptypage) {
	initpage ();
      }
      if (bofl) {
	if (offset > 0.0) {
	  fprintf (outfile, "%lf 0 rmoveto\n", offset * scale);
	}
	bofl = 0;
      }
      i=0;
      ptr = t;
      while (*t == ' ') {
	i++; t++;
      }
      if (prop == 1 && scs > 0.0) {
	if (kyoout) fprintf (outfile, "%% %d space characters\n", i);
	fprintf (outfile, "scs %d mul 0 rmoveto\n", i);
      }
      else {
	x = *t;
	*t = '\0';
	fprintf (outfile, "1 %d (%s) text\n", uline, outtext (ptr));
	*t = x;
      }
      break;

    default:
      if (emptypage) {
	initpage ();
      }
      if (bofl) {
	if (offset > 0.0) {
	  fprintf (outfile, "%lf 0 rmoveto\n", offset * scale);
	}
	bofl = 0;
      }
      ptr = t;
      while (*t && *t != '\033' &&
	     *t != '\r' && *t != '\n' &&
	     *t != ' '  && *t != '\010') {
	t++;
      }
      x = *t;
      *t = '\0';
      fprintf (outfile, "1 %d (%s) text\n", uline, outtext (ptr));
      *t = x;
      break;
    }
  }
}

void
yyerror (char *s)
{
  fprintf (stderr, "%s, line %d\n", s, lineno);
}

%}

%union {
  int 		v_keyword;
  struct {
    int move;
    int uline;
  } v_muopt;
  int		v_integer;
  double 	v_number;
  char 		v_string[256];
  rvrd         *v_rvrd;
}

%token <v_keyword>	PRESCRIBE EXIT SCRC SEM UNIT COPY CASS
%token <v_keyword>	STAK SSTK CSTK RES STAT FRPO SPSZ PAGE
%token <v_keyword>	TEXT RTXT CTXT CMNT STM SLM SRM SPW
%token <v_keyword>	SBM SPL SLPP SPO SLS SLPI SCS SCPI SULP
%token <v_keyword>	FONT ALTF SETF SCF RPF FLST FTMD SFA
%token <v_keyword>	INTL LDFC GENF DELF DAF MAP MZP MRP
%token <v_keyword>	MRPA SCP RPP SPD DAP DZP DRP DRPA BOX
%token <v_keyword>	CIR PIE FPAT BLK ARC STR RVRD ENDR
%token <v_keyword>	BARC MCRO ENDM CALL DELM DAM RDMP ENDD
%token <v_keyword>	XPAT PAT SRO SFNT FSET

%token <v_number>	NUMBER
%token <v_integer>	INTEGER
%token <v_string>	STRING OPTIONS PLAINTEXT COMMENT WS

%type <v_muopt>		muoption
%type <v_rvrd>		rvrd_data rvrd_line
%type <v_number>        number

%start start

%%

start:		file {{
	  if (kyoout) fprintf (outfile, "%% eof\n");
	  if (!emptypage) endpage ();
	}}
	|	/* nothing */
	;

file:		file data
	|	data
	;

data:		plaintext {}
	|	PRESCRIBE programm exit {}
	|	PRESCRIBE exit {}
	;

plaintext:	PLAINTEXT {{
	  const char * ptr = $1;
	  if (curpos.ypos > sbm) endpage ();
	  /* ignore if empty */
	  while (*ptr == ' ') ptr++;
	  if (*ptr || !emptypage) {
	    printplaintext ($1);
	  }
	}}
	|	'\n' {{
	  if (curpos.ypos > sbm) endpage ();
	  printplaintext ("\n");
	}}
	|	'\r' {{
	  printplaintext ("\r");
	}}
	|	'\f' {{
	  printplaintext ("\f");
	}}
	;

programm:	programm command ';'
	|	command ';'
	;

exit:		EXIT ';' {}
	|	EXIT ',' 'E' ';' {}
	|	EXIT ',' 'e' ';' {}
	;

command:	com_res		|	com_cmnt	|	com_unit
	|	com_page	|	com_stm		|	com_slm
	|	com_srm		|	com_spw		|	com_sbm
	|	com_spl		|	com_slpp	|	com_spo
	|	com_text	|	com_ctxt	|	com_rtxt
	|	com_map		|	com_mzp		|	com_mrp
	|	com_mrpa	|	com_scp		|	com_rpp
	|	com_spd		|	com_dap		|	com_dzp
	|	com_drp		|	com_drpa	|	com_box
	|	com_cir		|	com_fpat	|	com_blk
	|	com_font	|	com_ftmd	|	com_sls
	|	com_slpi	|	com_scs		|	com_scpi
	|	com_sulp	|	com_setf	|	com_altf
	|	com_scf		|	com_rpf		|	com_delm
	|	com_dam		|	com_arc		|	com_pie
	|	com_mcro	|	com_endm	|	com_call
	|	com_cass	|	com_stak	|	com_sstk
	|	com_cstk	|	com_spsz	|	com_copy
	|	com_frpo	|	com_ldfc	|	com_delf
	|	com_daf		|	com_sfa		|	com_str
	|	com_rvrd	|	com_xpat	|	com_pat
	|	com_intl	|	com_sro		|	com_sfnt
	|	com_fset
	|	/* empty */
	;

com_res:	RES {}
	;

com_cmnt:	CMNT cmnt_eatup {}
	|	CMNT {}
	;

cmnt_eatup:	cmnt_eatup COMMENT {{
	  fprintf (outfile, "%% %s\n", $2);
	}}
	|	COMMENT {{
	  fprintf (outfile, "%% %s\n", $1);
	}}
	;

com_unit:	UNIT OPTIONS {{
          if (kyoout) fprintf (outfile, "%% unit %s\n", $2);
	  switch ($2[0]) {
	  case 'I': case 'i': unit=UNIT_INCH; break;
	  case 'C': case 'c': unit=UNIT_CM;   break;
	  case 'P': case 'p': unit=UNIT_DOT;  break;
	  case 'D': case 'd': unit=UNIT_PIXEL;break;
	  }
	  umult = umumap[unit];
	}}
	;

com_page:	PAGE {{
	  if (kyoout) fprintf (outfile, "%% page\n");
	  if (!emptypage) endpage();
	}}
	;

com_stm:	STM NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% stm %f\n", $2);
	  if ($2 >= 0) {
	    stm = $2 * umult;
	    fprintf (outfile, "/stm %f def\n", stm * scale);
	  }
	  if (curpos.ypos < stm + sls) {
	    fprintf (outfile, "currentpoint pop stm sls add moveto\n");
	    curpos.ypos = stm + sls;
	  }
	}}
	;

com_slm:	SLM NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% slm %f\n", $2);
	  if ($2 >= 0) {
	    slm = $2 * umult;
	    fprintf (outfile, "/slm %f def\n", slm * scale);
	  }
	  if (curpos.xpos < slm) {
	    fprintf (outfile, "currentpoint exch pop slm exch moveto\n");
	    curpos.xpos = slm;
	  }
	}}
	;

com_srm:	SRM NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% srm %f\n", $2);
	  if ($2 >= 0 && $2*umult > slm) {
	    srm = $2 * umult;
	    fprintf (outfile, "/srm %f def\n", srm * scale);
	  }
	  if (srm > curpsize.xpos-unprintable.xpos) {
	    srm = curpsize.xpos - unprintable.xpos;
	    fprintf (outfile, "/srm %f def\n", srm * scale);
	  }
	}}
	;

com_spw:	SPW NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% spw %f\n", $2);
	  if ($2 >= 0) {
	    srm = $2 * umult + slm;
	    fprintf (outfile, "/srm %f def\n", srm * scale);
	  }
	  if (srm > curpsize.xpos-unprintable.xpos) {
	    srm = curpsize.xpos - unprintable.xpos;
	    fprintf (outfile, "/srm %f def\n", srm * scale);
	  }
	}}
	;

com_sbm:	SBM NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% sbm %f\n", $2);
	  if ($2 >= 0 && $2*umult > stm) {
	    sbm = $2 * umult;
	    fprintf (outfile, "/sbm %f def\n", sbm * scale);
	  }
	  if (sbm > curpsize.ypos-unprintable.ypos) {
	    sbm = curpsize.ypos - unprintable.ypos;
	    fprintf (outfile, "/sbm %f def\n", sbm * scale);
	  }
	}}
	;

com_spl:	SPL NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% spl %f\n", $2);
	  if ($2 >= 0) {
	    sbm = $2 * umult + stm;
	    fprintf (outfile, "/sbm %f def\n", sbm * scale);
	  }
	  if (sbm > curpsize.ypos-unprintable.ypos) {
	    sbm = curpsize.ypos - unprintable.ypos;
	    fprintf (outfile, "/sbm %f def\n", sbm * scale);
	  }
	}}
	;

com_slpp:	SLPP NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% slpp %f\n", $2);
	  if ($2 >= 0) {
	    sbm = $2 * sls + stm;
	    fprintf (outfile, "/sbm %f def\n", sbm * scale);
	  }
	  if (sbm > curpsize.ypos-unprintable.ypos) {
	    sbm = curpsize.ypos - unprintable.ypos;
	    fprintf (outfile, "/sbm %f def\n", sbm * scale);
	  }
	}}
	;

com_spo:	SPO OPTIONS {{
	  if (kyoout) fprintf (outfile, "%% spo %s\n", $2);
	  switch ($2[0]) {
	  case 'P': case 'p': set_spo (0, !emptypage, 0); break;
	  case 'L': case 'l': set_spo (1, !emptypage, 0); break;
	  }
	}}
	;

com_spsz:	SPSZ NUMBER {{
	  int size=(int)($2+0.5), oldspo=spo;

	  if (kyoout) { fprintf (outfile, "%% spsz %d\n", size); }

	  curpsize.xpos = psizes[size].xpos;
	  curpsize.ypos = psizes[size].ypos;

	  spo = -1;
	  set_spo (oldspo, !emptypage, 0);
	}}
	;

com_text:	TEXT muoption {{
	  if (curpos.ypos > sbm) endpage ();
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% text, %d\n", $2.move);
	  fprintf (outfile, "%d 0 0 kyomove\n", $2.move);
	  if ($2.move == 4 || $2.move == 5) curpos.ypos += sls;
	}}
	|	TEXT STRING muoption {{
	  if (curpos.ypos > sbm) endpage ();
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% text '%s', %d\n",
			       outtext($2), $3.move);
	  fprintf (outfile, "%d %d (%s) text\n",
		   $3.move, $3.uline, outtext($2));
	  if ($3.move == 4 || $3.move == 5) curpos.ypos += sls;
	}}
	;

com_ctxt:	CTXT muoption {{
	  if (curpos.ypos > sbm) endpage ();
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% ctxt, %d\n", $2.move);
	  fprintf (outfile, "%d 0 0 kyomove\n", $2.move);
	  if ($2.move == 4 || $2.move == 5) curpos.ypos += sls;
	}}
	|	CTXT STRING muoption {{
	  if (curpos.ypos > sbm) endpage ();
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% ctxt '%s', %d\n", 
			       outtext($2), $3.move);
	  fprintf (outfile, "%d %d (%s) ctxt\n",
		   $3.move, $3.uline, outtext($2));
	  if ($3.move == 4 || $3.move == 5) curpos.ypos += sls;
	}}
	;

com_rtxt:	RTXT muoption {{
	  if (curpos.ypos > sbm) endpage ();
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% rtxt, %d\n", $2.move);
	  fprintf (outfile, "%d 0 0 kyomove\n", $2.move);
	  if ($2.move == 4 || $2.move == 5) curpos.ypos += sls;
	}}
	|	RTXT STRING muoption {{
	  if (curpos.ypos > sbm) endpage ();
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% rtxt '%s', %d\n", 
			       outtext($2), $3.move);
	  fprintf (outfile, "%d %d (%s) rtxt\n",
		   $3.move, $3.uline, outtext($2));
	  if ($3.move == 4 || $3.move == 5) curpos.ypos += sls;
	}}
	;

muoption:	',' OPTIONS ',' OPTIONS {{
	  switch ($2[0]) {
	  case 'B': case 'b': $$.move = 0; break;
	  case 'H': case 'h': $$.move = 1; break;
	  case 'V': case 'v': $$.move = 2; break;
	  case 'E': case 'e': $$.move = 3; break;
	  case 'L': case 'l': $$.move = 4; break;
	  case 'N': case 'n': $$.move = 5; break;
	  default: $$.move = 0; break;
	  }
	  if ($4[0]=='u' || $4[0]=='U') 
	    $$.uline = 1;
	  else
	    $$.uline = 0;
	}}
	|	',' OPTIONS {{
	  switch ($2[0]) {
	  case 'B': case 'b': $$.move = 0; break;
	  case 'H': case 'h': $$.move = 1; break;
	  case 'V': case 'v': $$.move = 2; break;
	  case 'E': case 'e': $$.move = 3; break;
	  case 'L': case 'l': $$.move = 4; break;
	  case 'N': case 'n': $$.move = 5; break;
	  default: $$.move = 0; break;
	  }
	  $$.uline = 0;
	}}
	| /* empty */ {{
	  $$.move  = 0;
	  $$.uline = 0;
	}}
	;

com_map:	MAP NUMBER ',' NUMBER {{
	  if (kyoout) fprintf (outfile, "%% map %f, %f\n", $2, $4);
	  if (!emptypage) {
	    fprintf (outfile, "%f slm add %f stm add moveto\n",
		     $2 * umult * scale, $4 * umult * scale);
	  }
	  curpos.xpos = $2 * umult + slm;
	  curpos.ypos = $4 * umult + stm;
	}}
	;

com_mzp:	MZP NUMBER ',' NUMBER {{
	  if (kyoout) fprintf (outfile, "%% mzp %f, %f\n", $2, $4);
	  if (!emptypage) {
	    fprintf (outfile, "%f %f moveto\n",
		     $2 * umult * scale, $4 * umult * scale);
	  }
	  curpos.xpos = $2 * umult;
	  curpos.ypos = $4 * umult;
	}}
	;

com_mrp:	MRP NUMBER ',' NUMBER {{
	  if (kyoout) fprintf (outfile, "%% mrp %f, %f\n", $2, $4);
	  if (!emptypage) {
	    fprintf (outfile, "%f %f rmoveto\n",
		     $2 * umult * scale, $4 * umult * scale);
	  }
	  curpos.xpos += $2 * umult;
	  curpos.ypos += $4 * umult;
	}}
	;

com_mrpa:	MRPA NUMBER ',' NUMBER {{
	  double xtrans, ytrans;

	  xtrans = - sin ((- $4) * M_PI / 180.0) * $2 * umult;
	  ytrans =   cos ((- $4) * M_PI / 180.0) * $2 * umult;

	  if (spo == 0) ytrans = - ytrans;

	  if (kyoout) fprintf (outfile, "%% mrpa %f, %f\n", $2, $4);
	  if (!emptypage) {
	    fprintf (outfile, "%f %f rmoveto\n",
		     xtrans * scale, ytrans * scale);
	  }
	  curpos.xpos += xtrans;
	  curpos.ypos += ytrans;
	}}
	;

com_scp:	SCP {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% scp\n");
	  fprintf (outfile, "scp\n");
	  if (cpspos < 8) {
	    cpstack[cpspos].xpos = curpos.xpos;
	    cpstack[cpspos].ypos = curpos.ypos;
	    cpspos++;
	  }
	}}
	;

com_rpp:	RPP {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% rpp\n");
	  fprintf (outfile, "rpp\n");
	  if (cpspos > 0) {
	    cpspos--;
	    curpos.xpos = cpstack[cpspos].xpos;
	    curpos.ypos = cpstack[cpspos].ypos;
	  }
	}}
	;

com_spd:	SPD NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% spd %f\n", $2);
	  spd = $2 * umult;
	  fprintf (outfile, "%f setlinewidth\n", spd * scale);
	}}
	;

com_dap:	DAP NUMBER ',' NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% dap %f, %f\n", $2, $4);
	  fprintf (outfile, "currentpoint newpath moveto\n");
	  fprintf (outfile, "%f slm add %f stm add lineto\n",
		   $2 * umult * scale, $4 * umult * scale);
	  fprintf (outfile, "currentpoint stroke moveto\n");
	  curpos.xpos = $2 * umult + slm;
	  curpos.ypos = $2 * umult + stm;
	}}
	;

com_dzp:	DZP NUMBER ',' NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% dzp %f, %f\n", $2, $4);
	  fprintf (outfile, "currentpoint newpath moveto\n");
	  fprintf (outfile, "%f %f lineto\n",
		   $2 * umult * scale, $4 * umult * scale);
	  fprintf (outfile, "currentpoint stroke moveto\n");
	  curpos.xpos = $2 * umult;
	  curpos.ypos = $4 * umult;
	}}
	;

com_drp:	DRP NUMBER ',' NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% drp %f, %f\n", $2, $4);
	  fprintf (outfile, "currentpoint newpath moveto\n");
	  fprintf (outfile, "%f %f rlineto\n",
		   $2 * umult * scale, $4 * umult * scale);
	  fprintf (outfile, "currentpoint stroke moveto\n");
	  curpos.xpos += $2 * umult;
	  curpos.ypos += $4 * umult;
	}}
	;

com_drpa:	DRPA NUMBER ',' NUMBER {{
	  double xtrans, ytrans;

	  xtrans = - sin ((- $4) * M_PI / 180.0) * $2 * umult;
	  ytrans =   cos ((- $4) * M_PI / 180.0) * $2 * umult;

	  if (spo == 0) ytrans = -ytrans;

	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% drpa %f, %f\n", $2, $4);
	  fprintf (outfile, "currentpoint newpath moveto\n");
	  fprintf (outfile, "%f %f rlineto\n",
		   xtrans * scale, ytrans * scale);
	  fprintf (outfile, "currentpoint stroke moveto\n");

	  curpos.xpos += xtrans;
	  curpos.ypos += ytrans;
	}}
	;

com_box:	BOX NUMBER ',' NUMBER muoption {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% box %f, %f, %d\n", 
			       $2, $4, $5.move);
	  fprintf (outfile, "%d %f %f box\n",
		   $5.move,
		   $2 * umult * scale, $4 * umult * scale);
	}}
	;

com_cir:	CIR NUMBER {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% cir %f\n", $2);
	  fprintf (outfile, "%f cir\n",
		   $2 * umult * scale);
	}}
	;

com_arc:	ARC NUMBER ',' NUMBER ',' NUMBER ',' NUMBER {{
	  int i;
	  if (emptypage) initpage();
	  if (kyoout) fprintf (outfile, "%% arc %f, %f, %f, %f\n",
			       $2, $4, $6, $8);
	  for (i=0; i<32; i++) {
	    if (fpat[i] != 255 && fpat[i] != 0)
	      break;
	  }
	  if (i != 32) {
	    fprintf (stderr, "line %d: arc command with fill pattern not supported\n",
		     lineno);
	  }
	  else if (fpat[0] == 255) {
	    fprintf (outfile, "%f %f %f %f kyoarc\n",
		     $2 * umult * scale, $4 * umult * scale, $6, $8);
	  }
	  else {
	    fprintf (outfile, "%% arc with empty fill pattern ignored\n");
	  }
	}}
	;

com_pie:	PIE {{ piepieces = 0; }} NUMBER ',' NUMBER pie_slice {{
	  double pietotal=0.0, pieaccu, angle;
	  int index;

	  if (emptypage) initpage();
	  if (kyoout) {
	    fprintf (outfile, "%% pie %f, %f", $3, $5);
	    for (index=0; index<piepieces; index++) {
	      fprintf (outfile, ", %f", pieslices[index]);
	    }
	    fprintf (outfile, "\n");
	  }
	  for (index=0; index<piepieces; index++)
	    pietotal += pieslices[index];

	  fprintf (outfile, "%f cir\n", $3 * umult);
	  fprintf (outfile, "currentpoint newpath moveto\n");

	  for (index=0, pieaccu=0.0; index<piepieces; index++) {
	    angle = - (360.0*pieaccu/pietotal+$5) * M_PI / 180.0;
	    fprintf (outfile, "currentpoint %f %f rlineto moveto\n",
		     - sin (angle) * $3 * umult * scale,
		       cos (angle) * $3 * umult * scale * ((spo)?1:-1));
	    pieaccu += pieslices[index];
	  }
	  fprintf (outfile, "currentpoint stroke moveto\n");
	}}
	;

pie_slice:	pie_slice ',' NUMBER {{
	  pieslices[piepieces++] = $3;
	}}
	|	',' NUMBER {{
	  pieslices[piepieces++] = $2;
	}}
	;

com_fpat:	FPAT NUMBER ',' NUMBER ',' NUMBER ',' NUMBER ','
                     NUMBER ',' NUMBER ',' NUMBER ',' NUMBER {{
	  fpat[0]  = fpat[1]  = fpat[16] = fpat[17] = (unsigned char) ($2+.5);
	  fpat[2]  = fpat[3]  = fpat[18] = fpat[19] = (unsigned char) ($4+.5);
	  fpat[4]  = fpat[5]  = fpat[20] = fpat[21] = (unsigned char) ($6+.5);
	  fpat[6]  = fpat[7]  = fpat[22] = fpat[23] = (unsigned char) ($8+.5);
	  fpat[8]  = fpat[9]  = fpat[24] = fpat[25] = (unsigned char) ($10+.5);
	  fpat[10] = fpat[11] = fpat[26] = fpat[27] = (unsigned char) ($12+.5);
	  fpat[12] = fpat[13] = fpat[28] = fpat[29] = (unsigned char) ($14+.5);
	  fpat[14] = fpat[15] = fpat[30] = fpat[31] = (unsigned char) ($16+.5);
	  if (kyoout) {
	    fprintf (outfile, "%% fpat %d, %d, %d, %d, %d, %d, %d, %d\n",
		     (int) fpat[0], (int) fpat[2], (int) fpat[4], (int) fpat[6], 
		     (int) fpat[8], (int) fpat[10], (int) fpat[12], (int) fpat[14]);
	  }
	  if (!emptypage) {
	    kyofpat (fpat);
	  }
	}}
	;

com_blk:	BLK NUMBER ',' NUMBER muoption {{
	  if (emptypage) initpage ();
	  if (kyoout) fprintf (outfile, "%% blk %f, %f, %d\n", 
			       $2, $4, $5.move);
	  fprintf (outfile, "%d %f %f block\n",
		   $5.move,
		   $2 * umult * scale, $4 * umult * scale);
	}}
	;

com_font:	FONT NUMBER {{
	  font = (int) ($2+.5);

	  if (kyoout) fprintf (outfile, "%% font %d\n", font);
	  if (font < 1 || (font > MAXFONT && !havefont(font))) {
	    fprintf (outfile, "%% illegal font number %d, using 1\n", font);
	    fprintf (stderr,  "line %d: illegal font number %d, using 1\n",
		     lineno, font);
	    font = 1;
	  }
	  kyofont ();
	}}
	;

com_ftmd:	FTMD NUMBER {{
	  if (kyoout) fprintf (outfile, "%% ftmd %f\n", $2);
	  ftmd = (int)($2+.5);
	  if (!emptypage)
	    fprintf (outfile, "/ftmd %d def\n", ftmd);
	}}
	;

com_sls:	SLS NUMBER {{
	  if (kyoout) fprintf (outfile, "%% sls %f\n", $2);
	  sls = $2 * umult;
	  if (!emptypage)
	    fprintf (outfile, "/sls %f def\n", sls * scale);
	}}
	;

com_slpi:	SLPI NUMBER {{
	  if (kyoout) fprintf (outfile, "%% slpi %f\n", $2);
	  if ($2 > 0) 
	    sls = umumap[UNIT_INCH] / $2;
	  if (!emptypage)
	    fprintf (outfile, "/sls %f def\n", sls * scale);
	}}
	;

com_scs:	SCS NUMBER {{
	  if (kyoout) fprintf (outfile, "%% scs %f\n", $2);
	  if ($2*umult > 0.01)
	    scs = $2*umult;
	  else
	    scs = -1;
	  if (!emptypage)
	    fprintf (outfile, "/scs %f def\n", scs * scale);
	}}
	;

com_scpi:	SCPI NUMBER {{
	  if (kyoout) fprintf (outfile, "%% scpi %f\n", $2);
	  if ($2 > 0) {
	    scs = scale * umumap[UNIT_INCH] / $2;
	    if (!emptypage) {
	      fprintf (outfile, "/scs %f def\n", scs * scale);
	    }
	  }
	}}
	;

com_sulp:	SULP NUMBER ',' NUMBER {{
	  if (kyoout) fprintf (outfile, "%% sulp %f, %f\n", $2, $4);
	  if ($4 > 0) {
	    sulpd = $2 * umult;
	    sulps = $4 * umult;
	    if (!emptypage) {
	      fprintf (outfile, "/sulpd %f def /sulps %f def\n",
		       sulpd * scale, sulps * scale);
	    }
	  }
	}}
	;

com_setf:	SETF NUMBER ',' NUMBER {{
	  if (kyoout) fprintf (outfile, "%% setf %f, %f\n", $2, $4);
	  if ($2 >= 0 && $2 < 32 && $4 >= 0)
	    altf[(int)($2+.5)] = (int)($4+.5);
	}}
	;

com_altf:	ALTF NUMBER {{
	  if (kyoout) fprintf (outfile, "%% altf %f\n", $2);
	  if ($2 >= 0 && $2 < 32 && altf[(int)($2+.5)]>=0) {
	    font = altf[(int)($2+.5)];
	    kyofont ();
	  }
	}}
	;

com_scf:	SCF {{
	  if (kyoout) fprintf (outfile, "%% scf\n");
	  if (fontsp < 8)
	    fonts[fontsp++] = font;
	}}
	;

com_rpf:	RPF {{
	  if (kyoout) fprintf (outfile, "%% rpf\n");
	  if (fontsp > 0) {
	    font = fonts[--fontsp];
	    kyofont ();
	  }
	}}
	;

com_ldfc:	LDFC INTEGER ',' INTEGER ',' INTEGER ',' INTEGER ','
		     INTEGER ',' INTEGER ',' INTEGER ',' INTEGER ','
		     INTEGER ';' {{
	  if (kyoout) {
	    fprintf (outfile, "%% ldfc %d, %d, %d, %d, %d, %d, %d, %d, %d\n",
		     $2, $4, $6, $8, $10, $12, $14, $16, $18);
	  }

	  if ((newchar = (ldfc *) malloc (sizeof (ldfc))) == NULL) {
	    fprintf (stderr, "warning: out of mem in ldfc, cannot load character %d for font %d\n",
		     $4, $2);
	  } else {
	    newchar->fontnum   = $2;
	    newchar->code      = $4;
	    newchar->height    = $6;
	    newchar->width     = $8;
	    newchar->yoffset   = $10;
	    newchar->xoffset   = $12;
	    newchar->cellwidth = $14;
	    newchar->cellctr   = $16;
	    newchar->rotation  = $18;

	    if ((newchar->data = (unsigned char *) malloc (((newchar->width+15)/16)*2*newchar->height)) == NULL) {
	      fprintf (stderr, "warning: out of mem in ldfc, cannot load character %d for font %d\n",
		       $4, $2);
	      newchar->yi = -1;
	    } else {
	      memset (newchar->data, '\0', ((newchar->width+15)/16)*2*newchar->height);
	      newchar->xi = newchar->yi = 0;
	    }
	  }
	}} ldfc_pat {{
	  if (newchar && newchar->data) {
	    kyoldfc (newchar);
	  } else {
	    free (newchar);
	  }
	  newchar = NULL;
	}}
	;

ldfc_pat:	ldfc_pat OPTIONS {{
	  if (newchar && newchar->data && newchar->yi >= 0 && newchar->yi < newchar->height) {
	    unsigned char * ptr = newchar->data + 2*((newchar->width+15)/16)*newchar->yi+newchar->xi/8;
	    *ptr    =  (((($2[0])=='/')?127:$2[0]) - 64)         << 2;
	    *ptr++ |= ((((($2[1])=='/')?127:$2[1]) - 64) & 0x30) >> 4;
	    *ptr    = ((((($2[1])=='/')?127:$2[1]) - 64) & 0x0f) << 4;
	    *ptr   |=  (($2[2]) - 48) & 0x0f;
	    newchar->xi += 16;
	    if (newchar->xi >= newchar->width) {
	      newchar->yi += 1;
	      newchar->xi  = 0;
	    }
	  }
	}}
	|	OPTIONS {{
	  if (newchar && newchar->data && newchar->yi >= 0 && newchar->yi < newchar->height) {
	    unsigned char * ptr = newchar->data + 2*((newchar->width+15)/16)*newchar->yi+newchar->xi/8;
	    *ptr    =  (((($1[0])=='/')?127:$1[0]) - 64)         << 2;
	    *ptr++ |= ((((($1[1])=='/')?127:$1[1]) - 64) & 0x30) >> 4;
	    *ptr    = ((((($1[1])=='/')?127:$1[1]) - 64) & 0x0f) << 4;
	    *ptr   |=  (($1[2]) - 48) & 0x0f;
	    newchar->xi += 16;
	    if (newchar->xi >= newchar->width) {
	      newchar->yi += 1;
	      newchar->xi  = 0;
	    }
	  }
	}}
	;

com_delf:	DELF NUMBER {{
	  int tf = (int) ($2+.5);

	  if (kyoout) { fprintf (outfile, "%% delf %d\n", tf); }
	  kyodelf (tf);
	}}
	;

com_daf:	DAF {{
	  if (kyoout) { fprintf (outfile, "%% daf\n"); }
	  kyodaf ();
	}}
	;

com_sfa:	SFA NUMBER ',' NUMBER ',' OPTIONS {{
	  int fp = ($6[0] == 'F' || $6[0] == 'f') ? 0 : 1;
	  if (kyoout) { fprintf (outfile, "%% sfa %f, %f, %s\n",
				 $2, $4, $6); }
	  kyosfa  (font, $2*umult, $4*umult, fp);
	  kyofont ();
	}}
	;

com_str:	STR NUMBER {{
	  str = (int) ($2+.5);
	  if (kyoout) { fprintf (outfile, "%% str %d\n", str); }
	  if (!emptypage) fprintf (outfile, "/str %d def\n", str);
	}}
	;

com_rvrd:	RVRD rvrd_data ENDR {{
	  if (kyoout) { fprintf (outfile, "%% rvrd\n"); }
	  if ($2) {
	    kyorvrd ($2);
	  }
	  kyokillrd ($2);
	  if (kyoout) { fprintf (outfile, "%% endr\n"); }
	}}
	|	RVRD ENDR {{
	  if (kyoout) { fprintf (outfile, "%% rvrd\n"); }
	  if (kyoout) { fprintf (outfile, "%% endr\n"); }
	}}
	;

rvrd_data:	rvrd_data rvrd_line {{
	  $$ = $2;
	  $$->next = $1;
	}}
	|	rvrd_line {{
	  $$ = $1;
	}}
	;

rvrd_line:	INTEGER {{
	  if ((newrd = (rvrd *) malloc (sizeof (rvrd))) == NULL) {
	    fprintf (stderr, "warning: out of mem in rvrd, cannot load raster image\n");
	  }
	  else {
	    newrd->length = $1;
	    newrd->xi     = 0;
	    newrd->next   = NULL;
	    if ((newrd->data = (unsigned char *) malloc (newrd->length+1)) == NULL) {
	      fprintf (stderr, "warning: out of mem in rvrd, cannot load raster image\n");
	      free (newrd);
	      newrd = NULL;
	    } else {
	      memset (newrd->data, 0, newrd->length);
	    }
	  }
	}} rvrd_pat {{
	  $$ = newrd;
	}}
	;

rvrd_pat:	',' rvrd_ints ';'
	|	',' ';'
	|	';' rvrd_char ';'
	|	';' ';'
	;

rvrd_ints:	rvrd_ints ',' INTEGER {{
	  if (newrd && newrd->data && newrd->xi < newrd->length) {
	    newrd->data[newrd->xi] = (unsigned char) $3;
	  }
	  newrd->xi += 1;
	}}
	|	rvrd_ints ',' {{
	  if (newrd && newrd->data && newrd->xi < newrd->length) {
	    newrd->data[newrd->xi] = 0;
	  }
	  newrd->xi += 1;
	}}
	|	INTEGER {{
	  if (newrd && newrd->data && newrd->xi < newrd->length) {
	    newrd->data[newrd->xi] = (unsigned char) $1;
	  }
	  newrd->xi += 1;
	}}
	|	',' {{
	  if (newrd && newrd->data && newrd->xi < newrd->length) {
	    newrd->data[newrd->xi] = 0;
	  }
	  newrd->xi += 1;
	}}
	;

rvrd_char:	rvrd_char OPTIONS {{
	  if (newrd && newrd->data && newrd->xi < newrd->length) {
	    unsigned char * ptr = newrd->data + newrd->xi;
	    *ptr    =  (((($2[0])=='/')?127:$2[0]) - 64)         << 2;
	    *ptr++ |= ((((($2[1])=='/')?127:$2[1]) - 64) & 0x30) >> 4;
	    *ptr    = ((((($2[1])=='/')?127:$2[1]) - 64) & 0x0f) << 4;
	    *ptr   |=  (($2[2]) - 48) & 0x0f;
	    newrd->xi += 2;
	  }
	}}
	|	OPTIONS {{
	  if (newrd && newrd->data && newrd->xi < newrd->length) {
	    unsigned char * ptr = newrd->data + newrd->xi;
	    *ptr    =  (((($1[0])=='/')?127:$1[0]) - 64)         << 2;
	    *ptr++ |= ((((($1[1])=='/')?127:$1[1]) - 64) & 0x30) >> 4;
	    *ptr    = ((((($1[1])=='/')?127:$1[1]) - 64) & 0x0f) << 4;
	    *ptr   |=  (($1[2]) - 48) & 0x0f;
	    newrd->xi += 2;
	  }
	}}
	;

com_xpat:	XPAT INTEGER ';' {{
	  xpdef = $2; xpdi = 0;
	  if (kyoout) { fprintf (outfile, "%% xpat %d\n", xpdef); }
	  if (xpdef >= 100 && xpdef <= 105) {
	    memset (xpats[xpdef-100], 0, 32);
	  }
	}} xpat_data
	;

xpat_data:	xpat_data OPTIONS {{
	  if (xpdef >= 100 && xpdef <= 105 && xpdi >= 0 && xpdi < 32) {
	    unsigned char * ptr = xpats[xpdef-100] + xpdi;
	    *ptr    =  (((($2[0])=='/')?127:$2[0]) - 64)         << 2;
	    *ptr++ |= ((((($2[1])=='/')?127:$2[1]) - 64) & 0x30) >> 4;
	    *ptr    = ((((($2[1])=='/')?127:$2[1]) - 64) & 0x0f) << 4;
	    *ptr   |=  (($2[2]) - 48) & 0x0f;
	    xpdi += 2;
	  }
	}}
	|	OPTIONS {{
	  if (xpdef >= 100 && xpdef <= 105 && xpdi >= 0 && xpdi < 32) {
	    unsigned char * ptr = xpats[xpdef-100] + xpdi;
	    *ptr    =  (((($1[0])=='/')?127:$1[0]) - 64)         << 2;
	    *ptr++ |= ((((($1[1])=='/')?127:$1[1]) - 64) & 0x30) >> 4;
	    *ptr    = ((((($1[1])=='/')?127:$1[1]) - 64) & 0x0f) << 4;
	    *ptr   |=  (($1[2]) - 48) & 0x0f;
	  }
	    xpdi += 2;
	}}
	;

com_pat:	PAT NUMBER {{
	  int thepat = (int) ($2 + .5);
	  if (kyoout) { fprintf (outfile, "%% pat %d\n", thepat); }
	  if (thepat >= 1 && thepat <= 60) {
	    memcpy (fpat, prepats[thepat-1], 32);
	  }
	  else if (thepat >= 100 && thepat <= 105) {
	    memcpy (fpat, xpats[thepat-100], 32);
	  }
	  if (!emptypage) {
	    kyofpat (fpat);
	  }
	}}
	;

com_sro:	SRO NUMBER {{
	  if (kyoout) fprintf (outfile, "%% sro %f\n", $2);
	  sro = (int)($2+.5);
	  if (!emptypage)
	    fprintf (outfile, "/sro %d def\n", sro);
	}};

com_sfnt:	SFNT STRING {{
	  if (kyoout) fprintf (outfile, "%% sfnt %s\n", $2);
	  kyosfnt ($2, -1, -1, 0, 1, 0);
	}}
	|	SFNT STRING ',' number {{
	  if (kyoout) fprintf (outfile, "%% sfnt %s, %f\n", $2, $4);
	  kyosfnt ($2, $4, -1, 0, 1, 0);
	}}
	|	SFNT STRING ',' number ',' number {{
	  int fontno = (int) ($6+.5);
	  if (kyoout) fprintf (outfile, "%% sfnt %s, %f, %d\n",
			       $2, $4, fontno);
	  kyosfnt ($2, $4, fontno, 0, 1, 0);
	}}
	|	SFNT STRING ',' number ',' number ',' number {{
	  int fontno = (int) ($6+.5);
	  if (kyoout) fprintf (outfile, "%% sfnt %s, %f, %d, %f\n",
			       $2, $4, fontno, $8);
	  kyosfnt ($2, $4, fontno, 0, 1, 0);
	}}
	|	SFNT STRING ',' number ',' number ',' number ',' number {{
	  int fontno = (int) ($6+.5);
	  if (kyoout) fprintf (outfile, "%% sfnt %s, %f, %d, %f, %f\n",
			       $2, $4, fontno, $8, $10);
	  kyosfnt ($2, $4, fontno, 0, $10, 0);
	}}
	|	SFNT STRING ',' number ',' number ',' number ',' number
                            ',' number {{
	  int fontno = (int) ($6+.5);
	  if (kyoout) fprintf (outfile, "%% sfnt %s, %f, %d, %f, %f, %f\n",
			       $2, $4, fontno, $8, $10, $12);
	  kyosfnt ($2, $4, fontno, 0, $10, $12);
	}}
	;

com_fset:	FSET OPTIONS {{
	  char * ende, *ptr=$2;
	  ldfont * tf;
	  double val;
	  if (kyoout) fprintf (outfile, "%% fset %s\n", $2);
	  if (font == 0) {
	    tf = &curfont;
	  }
	  else {
	    tf = havefont (font);
	  }
	  if (tf == NULL || tf->type != SFNT_FONT) {
	    fprintf (stderr, "line %d: fset without sfnt ignored\n", lineno);
	  }
	  else {
	    while (val = strtod (ptr, &ende), *ende) {
	      switch (*ende) {
	      case 'V': case 'v':
		tf->sfnt.height = val;
		break;
	      case 'H': case 'h':
	      case 'P': case 'p':
	      case 'S': case 's':
	      case 'B': case 'b':
	      case 'T': case 't':
		fprintf (outfile, "%% fset option `%c' ignored\n", *ende);
		break;
	      default:
		fprintf (stderr, "line %d: illegal fset option `%c'\n",
			 lineno, *ende);
		break;
	      }
	      ptr = ende + 1;
	    }
	    kyofont ();
	  }
	}}
	;

com_delm:	DELM OPTIONS {{
	  if (kyoout) { fprintf (outfile, "%% delm %s ((ignored))\n", $2); }
	}}
	;

com_dam:	DAM {{
	  if (kyoout) { fprintf (outfile, "%% dam ((ignored))\n"); }
	}}
	;

com_cass:	CASS NUMBER {{
	  if (kyoout) { fprintf (outfile, "%% cass %d ((no effect))\n",
				 (int)($2+0.5)); }
	}}
	;

com_stak:	STAK NUMBER {{
	  if (kyoout) { fprintf (outfile, "%% stak %d ((no effect))\n",
				 (int)($2+0.5)); }
	}}
	;

com_sstk:	SSTK any_eatup {{
	  if (kyoout) { fprintf (outfile, "%% sstk ((no effect))\n"); }
	}}
	;

com_cstk:	CSTK any_eatup {{
	  if (kyoout) { fprintf (outfile, "%% cstk ((no effect))\n"); }
	}}
	;

com_copy:	COPY NUMBER {{
	  copies=(int)($2+0.5);
	  if (kyoout) { fprintf (outfile, "%% copy %d ((no effect))\n",
				 copies); }
	}}
	;

com_frpo:	FRPO OPTIONS NUMBER ',' NUMBER {{
	  int reg, val;
	  reg=(int)($3+0.5);
	  val=(int)($5+0.5);

	  if (kyoout) { fprintf (outfile, "%% frpo %s%d,%d ((no effect))\n",
				 $2, reg, val); }
	}}
	;

com_intl:	INTL NUMBER ',' NUMBER {{
	  int code, set;
	  code=(int)($2+0.5);
	  set=(int)($4+0.5);
	  if (kyoout) { fprintf (outfile, "%% intl %d, %d ((no effect))\n",
				 code, set); }
	}}
	;

com_mcro:	MCRO {{
	  fprintf (stderr, "line %d: error: macros must be preprocessed\n",
		   lineno);
	  /* parsing will fail at this point */
	}}
	;

com_endm:	ENDM {{
	  fprintf (stderr, "line %d: error: macros must be preprocessed\n",
		   lineno);
	  /* parsing will fail at this point */
	}}
	;

com_call:	CALL {{
	  fprintf (stderr, "line %d: error: macros must be preprocessed\n",
		   lineno);
	  /* parsing will fail at this point */
	}}
	;

number:		NUMBER
	|	INTEGER {{
	  $$ = $1;
	}}
	| /* empty */ {{
	  $$ = -1;
	}}
	;

any_eatup:	any_eatup OPTIONS
	|	any_eatup STRING
	|	any_eatup NUMBER
	|	any_eatup ','
	|	OPTIONS {}
	|	STRING {}
	|	NUMBER {}
	|	','
	;

%%
