/*
 * This file is part of the Prescribe to Postscript converter by
 * Frank Pilhofer <fp@informatik.uni-frankfurt.de>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * For non-commercial use ONLY.
 */

/*
 * Read X11 bitmaps and produce the predefined Kyocera Prescribe
 * patterns, whose bit orientation is different (X11: least sig-
 * nificant bit left, Prescribe: least significant bit right).
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "pat1"
#include "pat2"
#include "pat3"
#include "pat4"
#include "pat5"
#include "pat6"
#include "pat7"
#include "pat8"
#include "pat9"
#include "pat10"
#include "pat11"
#include "pat12"
#include "undef"

#define MAXPAT 60

char * p2p_makepats_c = "$Id: makepats.c,v 1.1 1998/03/08 18:15:45 fp Exp $";

static unsigned char *patbits[] = {
  pat1_bits,
  pat2_bits,
  pat3_bits,
  pat4_bits,
  pat5_bits,
  pat6_bits,
  pat7_bits,
  pat8_bits,
  pat9_bits,
  pat10_bits,
  pat11_bits,
  pat12_bits,
  NULL
};

unsigned char
reverse (unsigned char input)
{
  unsigned char datum = 0;

  if (input & (1<<0)) datum |= 1<<7;
  if (input & (1<<1)) datum |= 1<<6;
  if (input & (1<<2)) datum |= 1<<5;
  if (input & (1<<3)) datum |= 1<<4;
  if (input & (1<<4)) datum |= 1<<3;
  if (input & (1<<5)) datum |= 1<<2;
  if (input & (1<<6)) datum |= 1<<1;
  if (input & (1<<7)) datum |= 1<<0;

  return datum;
}

int
main (int argc, char *argv[])
{
  unsigned char **data = patbits;
  unsigned char *ptr;
  int i = 1, j;

  printf ("\
/*\n\
 * Predefined Prescribe patterns (PAT command)\n\
 *
 * Automatically generated -- do not edit\n\
 */\n\
\n\
char * p2p_patterns_c = \"$Id: makepats.c,v 1.1 1998/03/08 18:15:45 fp Exp $\";
\n\
unsigned char prepats[][32] = {\n\
");

  while (*data) {
    printf ("  /* %d */\n", i++);
    printf ("  {");
    for (j=0, ptr=*data; j<32; j++, ptr++) {
      if ((j%8)==0) {
	printf ("\n    ");
      }
      printf ("0x%02x%s", (int) reverse (*ptr), (j==31)?"":", ");
    }
    printf ("\n  }%s\n", (i<MAXPAT)?",":"");
    data++;
  }

  /*
   * set rest of patterns, if any, to "undef" pattern
   */

  for (; i<=MAXPAT; i++) {
    printf ("  /* %d */\n", i);
    printf ("  {");
    for (j=0, ptr=undef_bits; j<32; j++, ptr++) {
      if ((j%8)==0) {
	printf ("\n    ");
      }
      printf ("0x%02x%s", (int) reverse (*ptr), (j==31)?"":", ");
    }
    printf ("\n  }%s\n", (i<MAXPAT)?",":"");
  }
  printf ("};\n");
}
